BIBLIOGRAPHY_STR = """

@article{haghighat2021sciann,
  title={SciANN: A Keras/TensorFlow wrapper for scientific computations and physics-informed deep learning using artificial neural networks},
  author={Haghighat, Ehsan and Juanes, Ruben},
  journal={Computer Methods in Applied Mechanics and Engineering},
  volume={373},
  pages={113552},
  year={2021},
  publisher={Elsevier},
  url={https://www.sciencedirect.com/science/article/pii/S0045782520307374}
}

@article{raissi2019physics,
  title={Physics-informed neural networks: A deep learning framework for solving forward and inverse problems involving nonlinear partial differential equations},
  author={Raissi, Maziar and Perdikaris, Paris and Karniadakis, George E},
  journal={Journal of Computational Physics},
  volume={378},
  pages={686--707},
  year={2019},
  publisher={Elsevier},
  url={https://www.sciencedirect.com/science/article/pii/S0021999118307125}
}

@article{wang2020gp,
  title={Understanding and mitigating gradient pathologies in physics-informed neural networks},
  author={Wang, Sifan and Teng, Yujun and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2001.04536},
  year={2020},
  url={https://arxiv.org/abs/2001.04536}
}

@article{wang2020ntk,
  title={When and why PINNs fail to train: A neural tangent kernel perspective},
  author={Wang, Sifan and Yu, Xinling and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2007.14527},
  year={2020},
  url={https://arxiv.org/abs/2007.14527}
}

@article{li2017visualizing,
  title={Visualizing the loss landscape of neural nets},
  author={Li, Hao and Xu, Zheng and Taylor, Gavin and Studer, Christoph and Goldstein, Tom},
  journal={arXiv preprint arXiv:1712.09913},
  year={2017},
  url={https://arxiv.org/abs/1712.09913}
}

@article{wang2020eigenvector,
  title={On the eigenvector bias of Fourier feature networks: From regression to solving multi-scale PDEs with physics-informed neural networks},
  author={Wang, Sifan and Wang, Hanwen and Perdikaris, Paris},
  journal={arXiv preprint arXiv:2012.10047},
  year={2020},
  url={https://arxiv.org/abs/2012.10047}
}

@article{jagtap2020locally,
  title={Locally adaptive activation functions with slope recovery for deep and physics-informed neural networks},
  author={Jagtap, Ameya D and Kawaguchi, Kenji and Em Karniadakis, George},
  journal={Proceedings of the Royal Society A},
  volume={476},
  number={2239},
  pages={20200334},
  year={2020},
  publisher={The Royal Society},
  url={https://royalsocietypublishing.org/doi/abs/10.1098/rspa.2020.0334}
}

@article{sener2018multi,
  title={Multi-task learning as multi-objective optimization},
  author={Sener, Ozan and Koltun, Vladlen},
  journal={arXiv preprint arXiv:1810.04650},
  year={2018}
}

@inproceedings{chen2018gradnorm,
  title={Gradnorm: Gradient normalization for adaptive loss balancing in deep multitask networks},
  author={Chen, Zhao and Badrinarayanan, Vijay and Lee, Chen-Yu and Rabinovich, Andrew},
  booktitle={International Conference on Machine Learning},
  pages={794--803},
  year={2018},
  organization={PMLR}
}

@article{maddu2021inversedirichlet,
      title={Inverse-Dirichlet Weighting Enables Reliable Training of Physics Informed Neural Networks},
      author={Suryanarayana Maddu and Dominik Sturm and Christian L. Müller and Ivo F. Sbalzarini},
      year={2021},
      eprint={2107.00940},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}

@article{jagtap2021deep,
  title={Deep Kronecker neural networks: A general framework for neural networks with adaptive activation functions},
  author={Jagtap, Ameya D and Shin, Yeonjong and Kawaguchi, Kenji and Karniadakis, George Em},
  journal={arXiv preprint arXiv:2105.09513},
  year={2021}
}

@article{mcclenny2020self,
  title={Self-adaptive physics-informed neural networks using a soft attention mechanism},
  author={McClenny, Levi and Braga-Neto, Ulisses},
  journal={arXiv preprint arXiv:2009.04544},
  year={2020}
}

"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras.layers import InputLayer

from ..utils import to_list, floatx, set_floatx
from .rnn_functional import RNNFunctional


class RNNVariable(RNNFunctional):
    """ Configures the `LSTMVariable` object for the network's input.

    # Arguments
        units: Int.
            number of time units in an recurrent architecture.
            A minimum of 2 is needed.
        name: String.
            Required as derivatives work only with layer names.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').

    # Raises

    """
    def __init__(self,
                 units,
                 name=None,
                 dtype=None):

        if not dtype:
            dtype = floatx()
        elif not dtype == floatx():
            set_floatx(dtype)

        assert isinstance(units, int) and units>=2, \
            'RNN needs a minimum of 2 time units. '

        layer = InputLayer(
            batch_input_shape=(None, units, 1),
            name=name,
            dtype=dtype
        )

        super(RNNVariable, self).__init__(
            layers=to_list(layer),
            inputs=to_list(layer.input),
            outputs=to_list(layer.output),
        )

    @classmethod
    def get_class(cls):
        return RNNFunctional


import numpy as np
import matplotlib.pyplot as plt
import sciann as sn


omega = 2*np.pi/2.0
omega_bar = 2*np.pi/1.5
A = 1.0
beta = 1.0

tunits = 2
NDATA = 1000

t_data = np.linspace(0, 4*np.pi, NDATA).reshape(-1,1)
dt = np.diff(t_data.flatten()).mean()
t_data = t_data + np.linspace(0, dt*(tunits-1), tunits).reshape(1, -1)

y_data = A*(np.sin(omega*t_data) - beta*np.sin(omega_bar*t_data))

# Add noise
# y_noise = 0.15*np.std(y_data)*np.random.randn(NDATA)


t = sn.functionals.RNNVariable(tunits, 't')
y = sn.functionals.RNNFunctional('y', t, [1], 'sin', recurrent_activation='sin')

mRNN = sn.SciModel(t, y)

mRNN.train(t_data, y_data, learning_rate=0.001, epochs=10000, batch_size=100)




# -*- coding: utf-8 -*-
'''
General documentation architecture:

Home
Index

- Getting started
    Guide to SciANN model
    Guide to Functional
    FAQ

- SciModels
    About SciANN models
        explain when one should use Sequential or functional API
        explain compilation step
        explain weight saving, weight loading
        explain serialization, deserialization
    Sequential
    Model (functional API)

- Functionals
    About SciANN functionals
        explain common functional operations: get_weights, set_weights, get_config
        explain input_shape
        explain usage on non-Keras tensors
    Variables
    Fields
    Functionals 

Operations
Constraints 

'''
from sciann import utils
from sciann.functionals.mlp_functional import MLPFunctional
from sciann import Variable
from sciann import Field
from sciann import SciModel
from sciann import constraints
from sciann import utils
from sciann import Parameter


EXCLUDE = {
    'Constraint'
}

PAGES = [
    {
        'page': 'scimodels.md',
        'classes': [
            SciModel
        ],
        'methods': [
            SciModel.train,
            SciModel.predict,
            SciModel.loss_functions,
        ]
    },
    {
        'page': 'functionals.md',
        'classes': [
            MLPFunctional,
            Variable,
            Field,
            Parameter,
        ],
        'methods': [
            MLPFunctional.eval,
            MLPFunctional.get_weights,
            MLPFunctional.set_weights,
            MLPFunctional.reinitialize_weights,
            MLPFunctional.count_params,
            MLPFunctional.set_trainable,
            MLPFunctional.split,
        ]
    },
    {
        'page': 'variables.md',
        'classes': [
            Variable
        ]
    },
    {
        'page': 'fields.md',
        'classes': [
            Field
        ]
    },
    {
        'page': 'constraints.md',
        'classes': [
            constraints.Data,
            constraints.PDE,
            constraints.Tie
        ]
    },
    {
        'page': 'utils.md',
        'methods': [
            utils.math.grad,
            utils.math.diag_grad,
            utils.math.div,
            utils.math.radial_basis,
            utils.math.sin,
            utils.math.asin,
            utils.math.cos,
            utils.math.acos,
            utils.math.tan,
            utils.math.atan,
            utils.math.tanh,
            utils.math.exp,
            utils.math.pow,
            utils.math.add,
            utils.math.sub,
            utils.math.mul,
            utils.math.div,
        ],
    },
]

ROOT = 'https://sciann.com/'

template_np_implementation = """# Numpy implementation

    ```python
{{code}}
    ```
"""

template_hidden_np_implementation = """# Numpy implementation

    <details>
    <summary>Show the Numpy implementation</summary>

    ```python
{{code}}
    ```

    </details>
"""



""" Callbacks module define callbacks for training
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from multiprocessing import reduction

import tensorflow as tf
from tensorflow.python import keras as k
from keras import backend as K
from keras.callbacks import Callback
from tensorflow.python.util.tf_export import keras_export
from scipy.optimize import Bounds, LinearConstraint, minimize

import numpy as np

from .utilities import unpack_singleton, to_list, get_log_path
from .utilities import append_to_bib
from .math import tf_gradients
from time import time


@keras_export('keras.callbacks.EpochTime')
class EpochTime(Callback):
    """ Callback that records the training time.
    """
    def on_train_begin(self, logs={}):
        self.times = []

    def on_epoch_begin(self, batch, logs={}):
        self.epoch_time_start = time()

    def on_epoch_end(self, batch, logs={}):
        self.times.append(time() - self.epoch_time_start)
        logs['time'] = self.times[-1]


@keras_export('keras.callbacks.EarlyStoppingByLossVal')
class EarlyStoppingByLossVal(Callback):
    """ Callback that terminates the training when loss reaches a small values.
    """
    def __init__(self, value, stop_after=1):
        super(EarlyStoppingByLossVal, self).__init__()
        self.value = value
        self.wait = stop_after

    def on_epoch_end(self, epoch, logs={}):
        current = logs.get('loss')
        if current < self.value:
            self.wait -= 1
            if self.wait <= 0:
                self.model.stop_training = True
                print("Epoch {:05d}: early stopping at loss value {:0.6e}".format(epoch+1, current))
                print("Revise 'stop_loss_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))


@keras_export('keras.callbacks.EarlyStoppingByLearningRate')
class EarlyStoppingByLearningRate(Callback):
    """ Callback that terminates the training when learning rate reaches a small values.
    """
    def __init__(self, value):
        super(EarlyStoppingByLearningRate, self).__init__()
        self.value = value

    def on_epoch_end(self, epoch, logs={}):
        current = logs.get('lr')
        if current < self.value:
            self.model.stop_training = True
            print("Epoch {:05d}: early stopping at learning rate {:0.6e}".format(epoch+1, current))
            print("Revise 'stop_lr_value={:0.12f}' in '.train' if it was not your intent. ".format(self.value))


@keras_export('keras.callbacks.GradientPathologyLossWeight')
class GradientPathologyLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """
    def __init__(self, model, data_generator,
                 beta=0.1, freq=100, log_freq=None,
                 types=None,
                 alpha=0.0,
                 min_max=None,
                 **kwargs):
        super(GradientPathologyLossWeight, self).__init__()
        append_to_bib(["wang2020gp"])
        # generate samples. 
        self.inputs, self.targets, self.weights = data_generator[0]
        # eval loss and gradients.
        self.losses = []
        self.loss_grads = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            self.losses.append(K.function(model.inputs, f))
            gf = tf_gradients(f, model.trainable_weights, unconnected_gradients='zero')
            self.loss_grads.append([K.function(model.inputs, gf)])
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.beta = beta
        assert len(self.targets) == len(types)
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        if 'PDE' in self.types:
            self.eval_loss_weights = self.eval_loss_weights_m1
        else:
            self.eval_loss_weights = self.eval_loss_weights_m2
        self.alpha = alpha
        if alpha>0.:
            self.base_losses = [1. if li==0. else li for li in self.eval_losses()]
        else:
            self.base_losses = None
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max)==2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            self.update(epoch)

    def update(self, epoch):
        losses = self.eval_losses() if self.base_losses else None
        loss_gradients = self.eval_loss_gradients()
        self.update_loss_weights(epoch, loss_gradients, losses)
        self.update_loss_gradients(epoch, loss_gradients)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi
        # log norm2 of gradients
        for i, gi in enumerate(self.loss_gradients):
            logs[f'loss_gradient_{i}'] = gi
    
    def eval_losses(self):
        return [l(self.inputs) for l in self.losses]

    def eval_loss_gradients(self):
        # eval new gradients
        updated_grads = []
        for lgi in self.loss_grads:
            losses = []
            for lgij in lgi:
                losses.append(
                    np.concatenate(
                        [np.abs(wg).flatten() for wg in lgij(self.inputs)]
                    )
                )
            updated_grads.append(np.concatenate(losses))
        return updated_grads

    def eval_loss_weights_m1(self, updated_grads):
        # Method 1: normalization by PDE.
        # eval max normalization on PDE.
        ref_grad = []
        for type, ws in zip(self.types, updated_grads):
            if type == "PDE":
                ref_grad.append(ws.max())
        ref_grad = max(ref_grad)
        # mean loss terms
        mean_grad = []
        for type, ws in zip(self.types, updated_grads):
            # if type is not 'PDE':
            ws_mean = ws.mean()
            mean_grad += [1.0 if ws_mean == 0. else ws_mean]
        # mean_grad = np.mean(mean_grad)
        # evaluate new weights
        new_weights = []
        for i, type in enumerate(self.types):
            if type == 'PDE':
                new_weights.append(1.0)
            elif type in ('Data', 'Tie'):
                new_weights.append(ref_grad / mean_grad[i])
            else:
                assert 'Incorrect target. '
        return new_weights

    def eval_loss_weights_m2(self, updated_grads):
        # Method 1: normalization by PDE.
        # eval max normalization on PDE.
        ref_grad = []
        for ws in updated_grads:
            ref_grad.append(ws.max())
        normalization_grad = max(ref_grad)
        # mean loss terms
        mean_grad = [ws.mean() for ws in updated_grads]
        # evaluate new weights
        new_weights = []
        for i, type in enumerate(self.types):
            ref_grad_i = mean_grad[i] if mean_grad[i]>0. else np.mean(mean_grad)
            new_weights.append(normalization_grad / ref_grad_i)
        return new_weights

    def update_loss_weights(self, epoch, updated_grads, updated_losses):
        new_weights = self.eval_loss_weights(updated_grads)
        n_task = len(new_weights)
        
        if updated_losses:
            new_scores = [li/l0 for li, l0 in zip(updated_losses, self.base_losses)]
            new_scores = [(li/np.mean(new_scores))**self.alpha for li in new_scores]
        else:
            new_scores = n_task*[1.]

        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        gp_weights = [new_weights[i]/old_weights[i]*new_scores[i] for i in range(n_task)]
        # check for limiting weights.
        for i, wi in enumerate(gp_weights):
            if wi < self.min_max[0]:
                gp_weights[i] = self.min_max[0]
            elif wi > self.min_max[1]:
                gp_weights[i] = self.min_max[1]
        # normalizing weights.
        norm = len(gp_weights)/sum(gp_weights)
        gp_weights = [gpi*norm for gpi in gp_weights]

        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(old_weights):
            new_val = (1.0 - self.beta) * wi + self.beta * gp_weights[i]
            K.set_value(self.model.loss_weights[i], new_val)
            self.loss_weights.append(new_val)
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch+1), self.loss_weights)

    def update_loss_gradients(self, epoch, updated_grads):
        self.loss_gradients = [np.linalg.norm(lgi) for lgi in updated_grads]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'GP'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.GradNormLossWeight')
class GradNormLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the GradNorm algorithm by Chen et al.
    """
    def __init__(self, model, data_generator,
                 beta=0.1, freq=100, log_freq=None, types=None,
                 alpha=0.0, min_max=None, **kwargs):
        super(GradNormLossWeight, self).__init__()
        append_to_bib("chen2018gradnorm")
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]
        # eval loss and gradients.
        self.losses = []
        self.loss_grads = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            self.losses.append(K.function(model.inputs, f))
            gf = tf_gradients(f, model.trainable_weights, unconnected_gradients='zero')
            self.loss_grads.append([K.function(model.inputs, gf)])
        self.freq = np.Inf if (freq == False or freq == 0) else freq
        self.beta = beta
        assert len(self.targets) == len(types)
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        self.alpha = alpha
        if alpha > 0.:
            self.base_losses = [1. if li == 0. else li for li in self.eval_losses()]
        else:
            self.base_losses = None
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            self.update(epoch)

    def update(self, epoch):
        losses = self.eval_losses() if self.base_losses else None
        loss_gradients = self.eval_loss_gradients()
        self.update_loss_weights(epoch, loss_gradients, losses)
        self.update_loss_gradients(epoch, loss_gradients)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi
        # log norm2 of gradients
        for i, gi in enumerate(self.loss_gradients):
            logs[f'loss_gradient_{i}'] = gi

    def eval_losses(self):
        return [l(self.inputs) for l in self.losses]

    def eval_loss_gradients(self):
        # eval new gradients
        updated_grads = []
        for lgi in self.loss_grads:
            losses = []
            for lgij in lgi:
                losses.append(
                    np.concatenate(
                        [np.abs(wg).flatten() for wg in lgij(self.inputs)]
                    )
                )
            updated_grads.append(np.concatenate(losses))
        return updated_grads

    def eval_loss_weights(self, updated_grads):
        # get old weights
        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        # eval gradients.
        ref_grad = []
        for wi, gi in zip(old_weights, updated_grads):
            ref_grad.append(wi*np.linalg.norm(gi, 2))
        normalization_grad = np.mean(ref_grad)
        # evaluate new weights
        new_weights = []
        for i, type in enumerate(self.types):
            ref_grad_i = ref_grad[i]
            if ref_grad_i==0.:
                ref_grad_i = normalization_grad
            new_weights.append(normalization_grad / ref_grad_i)
        return new_weights

    def update_loss_weights(self, epoch, updated_grads, updated_losses):
        new_weights = self.eval_loss_weights(updated_grads)
        n_task = len(new_weights)

        if updated_losses:
            new_scores = [li / l0 for li, l0 in zip(updated_losses, self.base_losses)]
            mean_scores = np.mean(new_scores)
            new_scores = [(li / mean_scores) ** self.alpha for li in new_scores]
        else:
            new_scores = n_task * [1.]

        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        gp_weights = [new_weights[i] * new_scores[i] for i in range(n_task)]

        # check for limiting weights.
        for i, wi in enumerate(gp_weights):
            if wi < self.min_max[0]:
                gp_weights[i] = self.min_max[0]
            elif wi > self.min_max[1]:
                gp_weights[i] = self.min_max[1]

        # normalizing weights.
        norm = len(gp_weights) / sum(gp_weights)
        gp_weights = [gpi * norm for gpi in gp_weights]

        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(old_weights):
            new_val = (1.0 - self.beta) * wi + self.beta * gp_weights[i]
            K.set_value(self.model.loss_weights[i], new_val)
            self.loss_weights.append(new_val)
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)

    def update_loss_gradients(self, epoch, updated_grads):
        self.loss_gradients = [np.linalg.norm(lgi, 2) for lgi in updated_grads]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'GN'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.InverseDirichletLossWeight')
class InverseDirichletLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """

    def __init__(self, model, data_generator,
                 beta=0.5, freq=100, log_freq=None,
                 hessian=False, types=None,
                 alpha=0.0, min_max=None, **kwargs):
        super(InverseDirichletLossWeight, self).__init__()
        append_to_bib(["maddu2021inversedirichlet"])
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]
        # eval loss and gradients.
        self.losses = []
        self.loss_grads = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            self.losses.append(K.function(model.inputs, f))
            gf = tf_gradients(f, model.trainable_weights, unconnected_gradients='zero')
            if hessian is True:
                g2f = []
                for gfi in gf:
                    g2fi = tf_gradients(gfi, model.trainable_weights, unconnected_gradients='zero')
                    g2f.append(K.function(model.inputs, g2fi))
                self.loss_grads.append(g2f)
            else:
                self.loss_grads.append([K.function(model.inputs, gf)])
        self.freq = np.Inf if (freq == False or freq == 0) else freq
        self.beta = beta
        assert len(self.targets) == len(types)
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        self.alpha = alpha
        if alpha > 0.:
            self.base_losses = [1. if li==0. else li for li in self.eval_losses()]
        else:
            self.base_losses = None
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            self.update(epoch)

    def update(self, epoch):
        losses = self.eval_losses() if self.base_losses else None
        loss_gradients = self.eval_loss_gradients()
        self.update_loss_weights(epoch, loss_gradients, losses)
        self.update_loss_gradients(epoch, loss_gradients)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi
        # log norm2 of gradients
        for i, gi in enumerate(self.loss_gradients):
            logs[f'loss_gradient_{i}'] = gi

    def eval_losses(self):
        return [l(self.inputs) for l in self.losses]

    def eval_loss_gradients(self):
        # eval new gradients
        updated_grads = []
        for lgi in self.loss_grads:
            losses = []
            for lgij in lgi:
                losses.append(
                    np.concatenate(
                        [np.abs(wg).flatten() for wg in lgij(self.inputs)]
                    )
                )
            updated_grads.append(np.concatenate(losses))
        return updated_grads

    def eval_loss_weights(self, updated_grads):
        # Method 1: normalization by PDE.
        # eval max normalization on PDE.
        ref_grad = []
        for ws in updated_grads:
            ref_grad.append(ws.std())
        normalization_grad = max(ref_grad)
        # evaluate new weights
        new_weights = []
        for i, type in enumerate(self.types):
            ref_grad_i = ref_grad[i] if ref_grad[i]>0. else np.mean(ref_grad)
            new_weights.append(normalization_grad / ref_grad_i)
        return new_weights

    def update_loss_weights(self, epoch, updated_grads, updated_losses):
        new_weights = self.eval_loss_weights(updated_grads)
        n_task = len(new_weights)

        if updated_losses:
            new_scores = [li / l0 for li, l0 in zip(updated_losses, self.base_losses)]
            new_scores = [(li / np.mean(new_scores)) ** self.alpha for li in new_scores]
        else:
            new_scores = n_task * [1.]

        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        gp_weights = [new_weights[i] * new_scores[i] for i in range(n_task)]
        # check for limiting weights.
        for i, wi in enumerate(gp_weights):
            if wi < self.min_max[0]:
                gp_weights[i] = self.min_max[0]
            elif wi > self.min_max[1]:
                gp_weights[i] = self.min_max[1]
        # normalizing weights.
        norm = len(gp_weights) / sum(gp_weights)
        gp_weights = [gpi * norm for gpi in gp_weights]

        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(old_weights):
            new_val = (1.0 - self.beta) * wi + self.beta * gp_weights[i]
            K.set_value(self.model.loss_weights[i], new_val)
            self.loss_weights.append(new_val)
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)

    def update_loss_gradients(self, epoch, updated_grads):
        self.loss_gradients = [np.linalg.norm(lgi) for lgi in updated_grads]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'ID'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.SelfAdaptiveLossWeight')
class SelfAdaptiveLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """

    def __init__(self, model, data_generator,
                 eta=0.001, freq=1, log_freq=None,
                 min_max=None, **kwargs):
        super(SelfAdaptiveLossWeight, self).__init__()
        append_to_bib(["mcclenny2020self"])
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]
        # eval loss and gradients.
        self.losses = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            self.losses.append(K.function(model.inputs, f))
        self.freq = np.Inf if (freq == False or freq == 0) else freq
        self.eta = eta
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        self.base_losses = [1. if li==0. else li for li in self.eval_losses()]
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            self.update(epoch)

    def update(self, epoch):
        losses = self.eval_losses()
        self.update_loss_weights(epoch, losses)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi

    def eval_losses(self):
        return [l(self.inputs) for l in self.losses]

    def update_loss_weights(self, epoch, updated_losses):
        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        sa_weights = [wi + self.eta*li for wi, li in zip(old_weights, updated_losses)]
        # check for limiting weights.
        for i, wi in enumerate(sa_weights):
            if wi < self.min_max[0]:
                sa_weights[i] = self.min_max[0]
            elif wi > self.min_max[1]:
                sa_weights[i] = self.min_max[1]
        # normalizing weights.
        norm = len(sa_weights) / sum(sa_weights)
        sa_weights = [sai * norm for sai in sa_weights]

        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(sa_weights):
            K.set_value(self.model.loss_weights[i], wi)
            self.loss_weights.append(wi)
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'SALW'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.SelfAdaptiveSampleWeight')
class SelfAdaptiveSampleWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """

    def __init__(self, model, data_generator,
                 eta=0.001, freq=1, log_freq=None,
                 loss="MSE", d_mask_func=None,
                 min_max=None, **kwargs):
        super(SelfAdaptiveSampleWeight, self).__init__()
        append_to_bib(["mcclenny2020self"])
        # by default, a linear mask function is used. 
        if d_mask_func is None:
            self.d_mask_func = lambda x: 1
        else:
            self.d_mask_func = d_mask_func
        # generate samples.
        self.data_generator = data_generator
        if loss.lower() in ("mse", "meansquarederror", "mean_squared_error"):
            self.loss_func = lambda t, p: np.mean((t - p)**2, axis=-1)
        elif loss.lower() in ("mae", "meanabsoluteerror", "mean_absolute_error"):
            self.loss_func = lambda t, p: np.mean(abs(t - p), axis=-1)
        else:
            raise ValueError('Unrecognized loss function for self-adaptive-weights: ', loss)
        self.losses = []
        for i in range(len(model.outputs)):
            # fixed batch size
            loss_i = K.function(model.inputs, model.outputs[i])
            self.losses.append(loss_i)
        # eval loss and gradients.
        self.freq = np.Inf if (freq == False or freq == 0) else freq
        self.eta = eta
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            self.update(epoch)

    def update(self, epoch):
        losses = self.eval_losses()
        self.update_loss_weights(epoch, losses)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi

    def eval_losses(self):
        weights = self.data_generator._sample_weights
        losses = []
        for i, (loss, wi) in enumerate(zip(self.losses, weights)):
            ids = wi > 0
            inputs = [x[ids, :] for x in self.data_generator._inputs]
            trues = self.data_generator._outputs[i][ids]
            # eval loss
            loss_i = np.zeros_like(wi)
            loss_i[ids] = self.loss_func(trues, loss(inputs))
            losses.append(loss_i)
        return losses

    def update_loss_weights(self, epoch, updated_losses):
        sa_weights = []
        for w0, li in zip(self.data_generator._sample_weights, updated_losses):
            dmw = self.d_mask_func(w0)
            w1 = w0 + self.eta*dmw*li
            sa_weights.append(w1)
        # check for limiting weights.
        for i, wi in enumerate(sa_weights):
            sa_weights[i] = np.maximum(wi, self.min_max[0])
            sa_weights[i] = np.minimum(wi, self.min_max[1])
        # # normalize losses
        # scaling_factor = len(sa_weights) * sa_weights[0].size
        # sa_weights = [wi*(scaling_factor/np.sum(wi)) for wi in sa_weights]
        # update weights in data-generator
        self.data_generator._sample_weights = sa_weights
        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(sa_weights):
            self.loss_weights.append(wi.mean())
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'SASW'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs



@keras_export('keras.callbacks.SelfAdaptiveLossWeight2')
class SelfAdaptiveLossWeight2(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """

    def __init__(self, model, data_generator,
                 eta=0.001, freq=1, log_freq=None,
                 min_max=None, **kwargs):
        super(SelfAdaptiveLossWeight2, self).__init__()
        append_to_bib(["mcclenny2020self"])
        # generate samples.
        self.data_generator = data_generator
        self.model = model

    def on_epoch_begin(self, epoch, logs={}):
        self.data_generator.update_sample_weights()
        print('self-adaptive is updated')
        # pass

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'SA'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.CurriculumLossWeight')
class CurriculumLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """

    def __init__(self, model, data_generator,
                 freq=1, log_freq=None,
                 initial_weights=None,
                 final_weights=None,
                 curriculum_epochs=100,
                 delay_epochs=0,
                 **kwargs):
        super(CurriculumLossWeight, self).__init__()
        assert len(initial_weights) == len(model.outputs), "list of initial/final weights should be provided."
        assert len(final_weights) == len(model.outputs), "list of initial/final weights should be provided."
        assert curriculum_epochs > 0, "expecting a positive (>0) integer value for curriculum_epochs"
        self.model = model
        self.freq = np.Inf if (freq == False or freq == 0) else freq
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        self.initial_weights = initial_weights
        self.final_weights = final_weights
        self.learning_rate = [(wf-wi)/curriculum_epochs for wi, wf in zip (initial_weights, final_weights)]
        self.delay_epochs = delay_epochs
        self.curriculum_epochs = curriculum_epochs
        self.initialize_weights()

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0 and epoch >= self.delay_epochs:
            self.update(epoch)

    def update(self, epoch):
        self.update_loss_weights(epoch)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi

    def update_loss_weights(self, epoch):
        new_weights = []
        for i in range(len(self.model.loss_weights)):
            wi = self.initial_weights[i]
            lr = self.learning_rate[i]
            wf = min(wi + lr*(epoch - self.delay_epochs), self.final_weights[i])
            new_weights.append(wf)
        # evaluate new weights
        self.loss_weights = []
        for i, wi in enumerate(new_weights):
            K.set_value(self.model.loss_weights[i], wi)
            self.loss_weights.append(wi)
        # print updates
        print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)
    
    def initialize_weights(self):
        self.loss_weights = []
        for i, wi in enumerate(self.initial_weights):
            K.set_value(self.model.loss_weights[i], wi)
            self.loss_weights.append(wi)
        # print updates
        print('\n+ adaptive_weights is set to: ', self.loss_weights)

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'CLW'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.LossGradientHistory')
class LossGradientHistory(Callback):
    """ Callback that evaluate the gradient of loss terms.
    """
    def __init__(self, model, data_generator, path=None, freq=100, hessian=False):
        super(LossGradientHistory, self).__init__()
        # generate samples. 
        self.inputs, self.targets, self.weights = data_generator[0]

        self.loss_grads = []
        self.loss_hessians = [] if hessian else None
        for i in range(len(model.outputs)):
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            gf = tf_gradients(f, model.trainable_weights, unconnected_gradients='zero')
            self.loss_grads.append(
                K.function(model.inputs, gf)
            )
            if hessian:
                g2f = []
                for gfi in gf:
                    g2fi = tf_gradients(gfi, model.trainable_weights, unconnected_gradients='zero')
                    g2f.append(K.function(model.inputs, g2fi))
                self.loss_hessians.append(g2f)
        self.path = get_log_path(path, 'loss-')
        self.freq = np.Inf if (freq==False or freq==0) else freq

    def on_epoch_end(self, epoch, logs={}):
        loss_gradients = None
        if self.freq > 0 and epoch % self.freq == 0:
            loss_gradients = self.eval_loss_gradients()
            # output full gradient to the log path.
            for i, lg in enumerate(loss_gradients):
                np.savetxt(
                    self.path + "gradient-history-L{}-epoch{}.csv".format(i+1, epoch+1),
                    lg.reshape(-1, 1),
                    delimiter=','
                )
            # eval the hessian matrix.
            if self.loss_hessians:
                loss_hessians = self.eval_loss_hessians()
                for i, lh in enumerate(loss_hessians):
                    np.savetxt(
                        self.path + "hessian-history-L{}-epoch{}.csv".format(i+1, epoch+1),
                        lh.reshape(-1, 1),
                        delimiter=','
                    )

    def eval_loss_gradients(self):
        # eval new gradients
        updated_grads = []
        for lgi in self.loss_grads:
            updated_grads.append(
                np.concatenate(
                    [wg.flatten() for wg in lgi(self.inputs)]
                )
            )
        return updated_grads

    def eval_loss_hessians(self):
        # eval new gradients
        updated_hessians = []
        for lhi in self.loss_hessians:
            loss_hessians = []
            for lhij in lhi:
                wlhij = lhij(self.inputs)
                loss_hessians.append(
                    np.concatenate([wg.flatten() for wg in wlhij])
                )
            # append for each layer.
            updated_hessians.append(np.concatenate(loss_hessians))
        return updated_hessians

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.MTLLossWeight')
class MTLLossWeight(Callback):
    """ Callback that evaluate the gradient of loss terms.
    """

    def __init__(self, model, data_generator, freq=100, **kwargs):
        super(MTLLossWeight, self).__init__()
        append_to_bib("sener2018multi")
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]

        self.loss_grads = []
        self.parameters = model.trainable_weights
        for i in range(len(model.outputs)):
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            gf = tf_gradients(f, self.parameters, unconnected_gradients='zero')
            self.loss_grads.append(
                K.function(model.inputs, gf)
            )
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.hist_gradient = []
        self.hist_parameters = []
        self.loss_weights = len(model.outputs) * [1.]
        self.loss_hessians = len(model.outputs) * [0.]

    def on_epoch_begin(self, epoch, logs={}):
        # pass
        loss_gradients = None
        if epoch % self.freq == 0:
            grads = self.eval_loss_gradients()
            num_Ts = len(grads)

            M = np.array([[np.dot(gi, gj) for gj in grads] for gi in grads])
            alpha0 = np.ones(num_Ts)

            f = lambda x: 0.5*np.matmul(M, x).dot(x)
            jac = lambda x: np.matmul(M, x)
            hess = lambda x: M

            bounds = Bounds(np.zeros(num_Ts), np.ones(num_Ts)*num_Ts)
            constraint = LinearConstraint(np.ones(num_Ts), num_Ts, num_Ts)

            alpha = minimize(
                f, alpha0, method='trust-constr',
                jac=jac, hess=hess, bounds=bounds, constraints=constraint
            )

            self.loss_weights = alpha.x
            self.loss_gradients = [np.linalg.norm(hs) for hs in grads]

            for i, wi in enumerate(self.model.loss_weights):
                gp_weight = self.loss_weights[i]
                new_val = gp_weight  # (1.0 - self.beta) * K.get_value(wi) + self.beta * gp_weight
                K.set_value(self.model.loss_weights[i], new_val)

            # print updates
            print('\n+ adaptive_weights at epoch {}:'.format(epoch + 1), self.loss_weights)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        logs['adaptive_weights'] = self.loss_weights
        # log norm2 of gradients
        logs['loss_gradients'] = self.loss_gradients

    def eval_loss_gradients(self):
        # eval new gradients
        updated_grads = []
        for lgi in self.loss_grads:
            updated_grads.append(
                np.concatenate(
                    [wg.flatten() for wg in lgi(self.inputs)]
                )
            )
        return updated_grads

    def eval_parameters(self):
        return np.concatenate([K.get_value(wi).flatten() for wi in self.parameters])

    def eval_loss_hessians(self):
        # eval new gradients
        updated_hessians = []
        for lhi in self.loss_hessians:
            loss_hessians = []
            for lhij in lhi:
                wlhij = lhij(self.inputs)
                loss_hessians.append(
                    np.concatenate([wg.flatten() for wg in wlhij])
                )
            # append for each layer.
            updated_hessians.append(np.concatenate(loss_hessians))
        return updated_hessians

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.ScoreLossWeight')
class ScoreLossWeight(Callback):
    """ Callback that evaluate the gradient of loss terms.
    """

    def __init__(self, model, data_generator, freq=100, 
                 alpha=1., beta=0.1, min_max=None, **kwargs):
        super(ScoreLossWeight, self).__init__()
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]

        self.losses = []
        for i in range(len(model.outputs)):
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            f = tf.reduce_mean(
                model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            self.losses.append(K.function(model.inputs, f))

        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.alpha = alpha
        self.beta = beta
        self.base_losses = [1. if li == 0. else li for li in self.eval_losses()]
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    def on_epoch_begin(self, epoch, logs={}):
        # pass
        if epoch % self.freq == 0:
            losses = self.eval_losses()
            n_task = len(losses)

            scores = [li/l0 for li, l0 in zip(losses, self.base_losses)]
            scores = [(li/np.mean(scores))**self.alpha for li in scores]
            norm = n_task/sum(scores)
            new_weights = [si*norm for si in scores]

            # check for limiting weights.
            for i, wi in enumerate(new_weights):
                if wi < self.min_max[0]:
                    new_weights[i] = self.min_max[0]
                elif wi > self.min_max[1]:
                    new_weights[i] = self.min_max[1]

            old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
            
            self.loss_weights = []
            # evaluate new weights
            for i, wi in enumerate(old_weights):
                new_val = (1.0 - self.beta) * wi + self.beta * new_weights[i]
                K.set_value(self.model.loss_weights[i], new_val)
                self.loss_weights.append(new_val)

            # print updates
            print('\n+ adaptive_weights at epoch {}:'.format(epoch+1), self.loss_weights)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi

    def eval_losses(self):
        return [l(self.inputs) for l in self.losses]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.ParameterHistory')
class ParameterHistory(Callback):
    """ Callback that traces the evolution of parameters.
    """
    def __init__(self, parameters, freq=100):
        super(ParameterHistory, self).__init__()
        self.parameters = parameters
        self.freq = freq

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        if epoch % self.freq == 0:
            # log the weights in the history output.
            # logs['parameter_epochs'] = epoch
            for param in to_list(self.parameters):
                logs[param.name] = param.value

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1:
            kwargs['parameters'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.FunctionalHistory')
class FunctionalHistory(Callback):
    """ Callback that traces the evolution of field variables.
    """
    def __init__(self, functionals=[], inputs=None, sci_model=None, path=None, freq=100):
        super(FunctionalHistory, self).__init__()
        self.functionals = to_list(functionals)
        self.inputs = inputs
        self.sci_model = sci_model
        self.path = get_log_path(path, "functional-history")
        self.freq = freq

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        if epoch % self.freq == 0:
            logs['parameter_epochs'] = epoch
            self.log_functionals(epoch+1)

    def on_train_end(self, logs={}):
        self.log_functionals('-end')

    def log_functionals(self, epoch=None):
        # log the weights in the history output.
        for field in self.functionals:
            if self.sci_model is None:
                field_value = field.eval(self.inputs)
            else:
                field_value = field.eval(self.sci_model, self.inputs)
            np.savetxt(
                self.path + "-{}-epoch{}.csv".format(field.name, epoch),
                field_value,
                delimiter=','
            )

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1:
            kwargs['functionals'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


class LossLandscapeHistory(Callback):
    """ Callback that traces the evolution of loss function.
    """
    def __init__(self, model, data_generator,
                 norm=2, resolution=11, layer_wise=True,
                 path=None, trials=1):
        super(LossLandscapeHistory, self).__init__()
        append_to_bib("li2017visualizing")
        # generate samples. 
        inputs, targets, weights = data_generator[0]
        self._model = model
        self._inputs = inputs
        self._layers = [layer for layer in model.layers if layer.weights]
        self._weights_size = 0
        for layer in self._layers:
            for w in layer.weights:
                if not w.trainable:
                    continue
                self._weights_size += np.prod(w.shape)
        # collect initial value of the weights.
        self._norm = lambda xs: np.linalg.norm(xs, norm)
        self._resolution = resolution
        self._layer_wise = layer_wise
        self._weight_norm = [self._norm(np.concatenate(self._collect_weights()))]
        self._path = get_log_path(path, "loss-landscape-history")
        self._trials = trials

        f = None
        for i in range(len(model.outputs)):
            yp = model.outputs[i]
            ys = targets[i]
            ws = weights[i]
            fi = tf.reduce_mean(
                    model.loss_functions[i](ys, yp, sample_weight=ws)
            )
            if f is None:
                f = fi
            else:
                f += fi
        self._loss = K.function(model.inputs, f)
        self._loss_value = [self._loss(inputs)]

    def _update_weights(self, x):
        k = -1
        for layer in self._layers:
            w_list = []
            w_trainable = [w.trainable for w in layer.weights]
            batch_update = False not in w_trainable
            for w in layer.weights:
                if not w.trainable:
                    continue
                k += 1
                shape = w.get_shape()
                value = np.array(x[k]).reshape(shape)
                if batch_update:
                    w_list.append(value)
                else:
                    K.set_value(w, value)
            if batch_update:
                layer.set_weights(w_list)

    def _collect_weights(self):
        x_values = []
        for layer in self._layers:
            w_trainable = [w.trainable for w in layer.weights]
            for var, trainable in zip(layer.get_weights(), w_trainable):
                if trainable:
                    x_values.append(var.reshape(-1))
        return x_values

    def on_epoch_end(self, epoch, logs={}):
        # if epoch % self._freq != 0:
        #     return
        x_trained = self._collect_weights()
        self._weight_norm.append(self._norm(np.concatenate(x_trained)))
        self._loss_value.append(self._loss(self._inputs))
        logs['norm-loss-weights'] = self._weight_norm[-1]

    def on_train_end(self, logs={}):
        x_trained = self._collect_weights()
        x_sizes = [x.size for x in x_trained]
        num_param = sum(x_sizes)
        # different trials to check consistency of landscape plot.
        for trial in range(self._trials):
            n0 = np.split(np.random.standard_normal(num_param), np.cumsum(x_sizes))
            n1 = np.split(np.random.standard_normal(num_param), np.cumsum(x_sizes))
            # normalize layer wise
            if self._layer_wise:
                n0 = [ni/self._norm(ni) for ni in n0]
                n1 = [ni/self._norm(ni) for ni in n1]
            # normalize globally
            n0_norm, n1_norm = [self._norm(np.concatenate(ni)) for ni in [n0, n1]]
            n0 = [ni/n0_norm for ni in n0]
            n1 = [ni/n1_norm for ni in n1]

            delta_weights = 2.0 * abs(self._weight_norm[-1] - self._weight_norm[0])
            loss_values = np.zeros((self._resolution**2, 3))

            k = 0
            for i, l0 in enumerate(np.linspace(-delta_weights, delta_weights, self._resolution)):
                # l0_weight = self._norm(np.concatenate([xi + n0i*l0 for xi,n0i in zip(x_trained, n0)]))
                for j, l1 in enumerate(np.linspace(-delta_weights, delta_weights, self._resolution)):
                    # l1_weight = self._norm(np.concatenate([xi + n1i*l1 for xi,n1i in zip(x_trained, n1)]))
                    test_weights = [xi + n0i*l0 + n1i*l1 for xi,n0i,n1i in zip(x_trained, n0, n1)]
                    self._update_weights(test_weights)
                    loss_values[k, :] = [l0, l1, self._loss(self._inputs)]
                    k += 1

            # save the calculations
            np.savetxt(
                # self._path + "-epoch{}.csv".format(epoch + 1),
                self._path + "-landscape{}.csv".format("" if self._trials==1 else trial),
                loss_values,
                delimiter=','
            )

        # save weight history
        np.savetxt(
            self._path + "-trace.csv",
            [[n0i, n0i, li] for n0i, li in zip(self._weight_norm, self._loss_value)],
            delimiter=','
        )
        # reset the weights
        self._update_weights(x_trained)

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        if len(args) == 1 and isinstance(args[0], int):
            kwargs['resolution'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs



@keras_export('keras.callbacks.NTKLossWeight')
class NTKLossWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """
    def __init__(self, model, data_generator,
                 beta=0.9, freq=100, log_freq=None,
                 alpha=0., types=None, min_max=None, **kwargs):
        super(NTKLossWeight, self).__init__()
        # print(f"{beta = }, {freq = }, {alpha = }, {min_max = }")
        append_to_bib("wang2020ntk")
        # generate samples.
        self.inputs, self.targets, self.weights = data_generator[0]
        # eval loss and gradients.
        self.losses, self.loss_grads = [], []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            # eval losses
            if alpha>0.:
                f = tf.reduce_mean(
                    model.loss_functions[i](ys, yp, sample_weight=ws)
                )
                self.losses.append(K.function(model.inputs, f))
            # eval gradients
            gf = tf_gradients(yp, model.trainable_weights, unconnected_gradients='zero')
            gf_dot_gf = tf.add_n([tf.reduce_sum(tf.multiply(gf_i, gf_i)) for gf_i in gf])
            self.loss_grads.append(K.function(model.inputs, gf_dot_gf))
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.beta = beta
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        assert len(model.outputs) == len(types)
        self.alpha = alpha
        if alpha>0.:
            self.base_losses = self.eval_losses()
        else:
            self.base_losses = None
        if min_max is None:
            self.min_max = [-np.Inf, np.Inf]
        else:
            assert len(min_max) == 2
            self.min_max = min_max

    @staticmethod
    def collect_weights(model):
        x_values = []
        for weights in model.trainable_weights:
            x_values.append(tf.reshape(weights, (-1,)))
        return tf.concat(x_values, axis=0)

    def on_epoch_begin(self, epoch, logs={}):
        if epoch % self.freq == 0:
            losses = self.eval_losses() if self.base_losses else None
            loss_gradients = self.eval_diag_ntk()
            self.update_loss_weights(epoch, loss_gradients, losses)
            self.update_loss_gradients(epoch, loss_gradients)

    def on_epoch_end(self, epoch, logs={}):
        # log gradient values
        for i, wi in enumerate(self.loss_weights):
            logs[f'loss_weight_{i}'] = wi
        # log norm2 of gradients
        for i, gi in enumerate(self.loss_gradients):
            logs[f'loss_gradient_{i}'] = gi

    def eval_diag_ntk(self):
        # eval new gradients
        num_samples = self.inputs[0].shape[0]
        updated_NTK = [np.zeros((num_samples,)) for lgi in self.loss_grads]
        start_time = time()
        for li, lgi in enumerate(self.loss_grads):
            for i in range(num_samples):
                wi = self.weights[li][i]
                if wi == 0.: continue
                xi = [v[i:i + 1, :] for v in self.inputs]
                updated_NTK[li][i] = lgi(xi)
        print("\n+ NTK evaluation time [s]: {} ".format(time() - start_time))
        return updated_NTK

    def eval_losses(self):
        losses = [fi(self.inputs) for fi in self.losses]
        return [1. if li==0. else li for li in losses]

    def eval_loss_weights(self, updated_grads):
        # eval max normalization on PDE.
        ref_grad = []
        for ws in updated_grads:
            ref_grad.append(ws.sum())
        normalization_grad = sum(ref_grad)
        # evaluate new weights
        new_weights = []
        for i, type in enumerate(self.types):
            ref_grad_i = ref_grad[i] if ref_grad[i] != 0. else np.mean(ref_grad)
            new_weights.append(normalization_grad / ref_grad_i)
        return new_weights

    def update_loss_weights(self, epoch, updated_grads, updated_losses):
        new_weights = self.eval_loss_weights(updated_grads)
        # print(f"{new_weights = }")
        n_task = len(new_weights)

        if updated_losses:
            new_scores = [li / l0 for li, l0 in zip(updated_losses, self.base_losses)]
            mean_scores = np.mean(new_scores)
            new_scores = [(li / mean_scores) ** self.alpha for li in new_scores]
        else:
            new_scores = n_task * [1.]
        # print(f"{self.base_losses = }")
        # print(f"{updated_losses = }")
        # print(f"{new_scores = }")

        old_weights = [K.get_value(wi) for wi in self.model.loss_weights]
        ntk_weights = [new_weights[i] * new_scores[i] for i in range(n_task)]
        # print(f"{ntk_weights = }")

        # check for limiting weights.
        for i, wi in enumerate(ntk_weights):
            if wi < self.min_max[0]:
                ntk_weights[i] = self.min_max[0]
            elif wi > self.min_max[1]:
                ntk_weights[i] = self.min_max[1]

        # normalizing weights.
        norm = n_task / sum(ntk_weights)
        ntk_weights = [wi * norm for wi in ntk_weights]

        # print(f"{norm = }")
        # print(f"{ntk_weights = }")

        self.loss_weights = []
        # evaluate new weights
        for i, wi in enumerate(old_weights):
            new_val = (1-self.beta) * wi + self.beta * ntk_weights[i]
            K.set_value(self.model.loss_weights[i], new_val)
            self.loss_weights.append(new_val)

        # print updates
        print(f'\n+ adaptive_weights at epoch {epoch+1}:', self.loss_weights)

    def update_loss_gradients(self, epoch, updated_grads):
        self.loss_gradients = [np.linalg.norm(lgi, 2) for lgi in updated_grads]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'NTK'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


#
# @keras_export('keras.callbacks.NTKLossWeight')
# class NTKLossWeight(Callback):
#     """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
#     """
#     def __init__(self, model, data_generator,
#                  beta=0.1, freq=100, log_freq=None,
#                  hessian=False, types=None, **kwargs):
#         super(NTKLossWeight, self).__init__()
#         append_to_bib("wang2020ntk")
#         # generate samples.
#         inputs, targets, weights = data_generator.get_data()
#         # limit number of samples for performance concerns.
#         MAX_SAMPLE = 10000
#         if inputs[0].shape[0] > MAX_SAMPLE:
#             sample_ids = np.random.choice(inputs[0].shape[0], MAX_SAMPLE, replace=False)
#             self.inputs = [x[sample_ids] for x in inputs]
#             self.targets = [y[sample_ids] for y in targets]
#             self.weights = [w[sample_ids] for w in weights]
#         else:
#             self.inputs = inputs
#             self.targets = targets
#             self.weights = weights
#         # eval loss and gradients.
#         self.loss_grads = []
#         for i in range(len(model.outputs)):
#             # fixed batch size
#             yp = model.outputs[i]
#             ws = self.weights[i]
#             gf = tf_gradients(yp, model.trainable_weights, unconnected_gradients='zero')
#             self.loss_grads.append(K.function(model.inputs, gf))
#         self.freq = np.Inf if (freq==False or freq==0) else freq
#         self.beta = beta
#         self.types = types
#         if log_freq is None:
#             self.log_freq = self.freq
#         else:
#             self.log_freq = log_freq
#         assert len(model.outputs) == len(types)
#
#     @staticmethod
#     def collect_weights(model):
#         x_values = []
#         for weights in model.trainable_weights:
#             x_values.append(tf.reshape(weights, (-1,)))
#         return tf.concat(x_values, axis=0)
#
#     def update(self, epoch):
#         loss_gradients = self.eval_diag_ntk()
#         self.update_loss_weights(epoch, loss_gradients)
#         self.update_loss_gradients(epoch, loss_gradients)
#
#     def on_train_begin(self, logs={}):
#         self.update(0)
#
#     def on_epoch_begin(self, epoch, logs={}):
#         if self.freq > 0 and (epoch + 1) % self.freq == 0:
#             self.update(epoch)
#
#     def on_epoch_end(self, epoch, logs={}):
#         # log gradient values
#         logs['adaptive_weights'] = self.loss_weights
#         # log norm2 of gradients
#         logs['loss_gradients'] = self.loss_gradients
#
#     def eval_full_ntk(self):
#         # eval new gradients
#         num_samples = self.inputs[0].shape[0]
#         ntk = [
#             [np.zeros((num_samples,num_samples)) for lgi in self.loss_grads]
#             for lgj in self.loss_grads
#         ]
#         for i in range(num_samples):
#             xi = [v[i:i+1, :] for v in self.inputs]
#             for j in range(num_samples):
#                 xj = [v[j:j+1, :] for v in self.inputs]
#                 for li, lgi in enumerate(self.loss_grads):
#                     if self.weights[li][i] == 0.: continue
#                     gi = np.concatenate([np.abs(wg).flatten() for wg in lgi(xi)])
#                     for lj, lgj in enumerate(self.loss_grads):
#                         if self.weights[lj][j] == 0.: continue
#                         gj = np.concatenate([np.abs(wg).flatten() for wg in lgj(xj)])
#                         gij = np.dot(gi, gj)
#                         ntk[li][lj][i, j] = gij
#         return ntk
#
#     def eval_diag_ntk(self):
#         # eval new gradients
#         num_samples = self.inputs[0].shape[0]
#         updated_NTK = [np.zeros((num_samples,)) for lgi in self.loss_grads]
#         # start_time = time()
#         for li, lgi in enumerate(self.loss_grads):
#             for i in range(num_samples):
#                 wi = self.weights[li][i]
#                 if wi == 0.: continue
#                 xi = [v[i:i + 1, :] for v in self.inputs]
#                 gi = np.concatenate([np.abs(wg).flatten() for wg in lgi(xi)])
#                 gij = np.dot(gi, gi)
#                 updated_NTK[li][i] = gij
#         # print(time() - start_time)
#         return updated_NTK
#
#     def eval_loss_weights(self, updated_grads):
#         # Method 1: normalization by PDE.
#         # eval max normalization on PDE.
#         ref_grad = []
#         for ws in updated_grads:
#             ref_grad.append(ws.sum())
#         normalization_grad = sum(ref_grad)
#         # evaluate new weights
#         new_weights = []
#         for i, type in enumerate(self.types):
#             ref_grad_i = ref_grad[i] if ref_grad[i] != 0. else np.mean(ref_grad)
#             new_weights.append(normalization_grad / ref_grad_i)
#         return new_weights
#
#     def update_loss_weights(self, epoch, updated_grads):
#         new_weights = self.eval_loss_weights(updated_grads)
#         norm = 1.0 #sum(new_weights)
#         self.loss_weights = []
#         # evaluate new weights
#         for i, wi in enumerate(self.model.loss_weights):
#             # gp_weight = new_weights[i] / K.get_value(wi)
#             # new_val = (1.0 - self.beta) * K.get_value(wi) + self.beta * gp_weight
#             new_val = new_weights[i]/norm
#             K.set_value(self.model.loss_weights[i], new_val)
#             self.loss_weights.append(new_val)
#         # print updates
#         print('\n+ adaptive_weights at epoch {}:'.format(epoch+1), self.loss_weights)
#
#     def update_loss_gradients(self, epoch, updated_grads):
#         self.loss_gradients = [np.linalg.norm(lgi) for lgi in updated_grads]
#
#     @staticmethod
#     def prepare_inputs(*args, **kwargs):
#         kwargs['method'] = 'NTK'
#         if len(args) == 1:
#             kwargs['freq'] = args[0]
#         elif len(args) > 0:
#             raise ValueError
#         return kwargs
#

@keras_export('keras.callbacks.NTKSampleWeight')
class NTKSampleWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """
    def __init__(self, model, data_generator,
                 beta=0.1, freq=100, log_freq=None,
                 hessian=False, types=None, **kwargs):
        super(NTKSampleWeight, self).__init__()
        # generate samples. 
        inputs, targets, weights = data_generator.get_data()
        # limit number of samples for performance concerns.
        self.inputs = inputs
        self.targets = targets
        self.weights = weights
        # eval loss and gradients.
        self.loss_grads = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ws = self.weights[i]
            gf = tf_gradients(yp, model.trainable_weights, unconnected_gradients='zero')
            self.loss_grads.append(K.function(model.inputs, gf))
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.beta = beta
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        assert len(model.outputs) == len(types)

    @staticmethod
    def collect_weights(model):
        x_values = []
        for weights in model.trainable_weights:
            x_values.append(tf.reshape(weights, (-1,)))
        return tf.concat(x_values, axis=0)

    def update(self, epoch):
        loss_gradients = self.eval_diag_ntk()
        self.update_loss_weights(epoch, loss_gradients)
        self.update_loss_gradients(epoch, loss_gradients)

    def on_train_begin(self, logs={}):
        self.update(0)

    def on_epoch_begin(self, epoch, logs={}):
        if self.freq > 0 and (epoch + 1) % self.freq == 0:
            self.update(epoch)

    def on_epoch_end(self, epoch, logs={}):
        pass
        # # log gradient values
        # logs['adaptive_weights'] = self.loss_weights
        # # log norm2 of gradients
        # logs['loss_gradients'] = self.loss_gradients

    def eval_full_ntk(self):
        # eval new gradients
        num_samples = self.inputs[0].shape[0]
        ntk = [
            [np.zeros((num_samples,num_samples)) for lgi in self.loss_grads]
            for lgj in self.loss_grads
        ]
        for i in range(num_samples):
            xi = [v[i:i+1, :] for v in self.inputs]
            for j in range(num_samples):
                xj = [v[j:j+1, :] for v in self.inputs]
                for li, lgi in enumerate(self.loss_grads):
                    gi = np.concatenate([np.abs(wg).flatten() for wg in lgi(xi)])
                    for lj, lgj in enumerate(self.loss_grads):
                        gj = np.concatenate([np.abs(wg).flatten() for wg in lgj(xj)])
                        gij = np.dot(gi, gj)
                        ntk[li][lj][i, j] = gij
        return ntk

    def eval_diag_ntk(self):
        # eval new gradients
        num_samples = self.inputs[0].shape[0]
        updated_NTK = [np.zeros((num_samples,)) for lgi in self.loss_grads]
        for li, lgi in enumerate(self.loss_grads):
            for i in range(num_samples):
                xi = [v[i:i + 1, :] for v in self.inputs]
                gi = np.concatenate([np.abs(wg).flatten() for wg in lgi(xi)])
                gij = np.dot(gi, gi)
                updated_NTK[li][i] = gij
        return updated_NTK

    def eval_sample_weights(self, updated_grads):
        # Method 1: normalization by PDE.
        # evaluate new weights
        new_weights = []
        tr_ntk = np.sum([np.sum(wi) for wi in updated_grads])
        for i, ws in enumerate(updated_grads):
            ws[ws != 0.] = tr_ntk / ws[ws != 0.]
            new_weights.append(ws)
        return new_weights

    def update_loss_weights(self, epoch, updated_grads):
        new_weights = self.eval_sample_weights(updated_grads)
        self.sample_weights = []
        # evaluate new weights
        for i, wi in enumerate(self.model.sample_weights):
            # gp_weight = new_weights[i] / K.get_value(wi)
            # new_val = (1.0 - self.beta) * K.get_value(wi) + self.beta * gp_weight
            new_val = new_weights[i]
            self.weights[i][:] = new_val

    def update_loss_gradients(self, epoch, updated_grads):
        self.loss_gradients = [np.linalg.norm(lgi) for lgi in updated_grads]

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'NTK'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.AdaptiveSampleWeight')
class AdaptiveSampleWeight(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """
    def __init__(self, model, data_generator,
                 beta=0.1, freq=100, log_freq=None,
                 hessian=False, types=None, **kwargs):
        super(AdaptiveSampleWeight, self).__init__()
        # generate samples. 
        inputs, targets, weights = data_generator.get_grid()
        # limit number of samples for performance concerns.
        self.inputs = inputs
        self.targets = targets
        self.weights = weights
        # eval loss and gradients.
        self.losses = []
        for i in range(len(model.outputs)):
            # fixed batch size
            yp = model.outputs[i]
            ys = self.targets[i]
            ws = self.weights[i]
            self.losses.append(
                K.function(model.inputs, yp)
            )
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.beta = beta
        self.types = types
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq
        assert len(model.outputs) == len(types)

    def on_train_begin(self, logs={}):
        pass

    def on_epoch_begin(self, epoch, logs={}):
        if (epoch + 1) % self.freq == 0:
            for i, loss in enumerate(self.losses):
                wi = np.abs(loss(self.inputs) - self.targets[i]).flatten()
                wi[wi != 0.] = 1. / wi[wi != 0.]
                wi *= (wi.size / np.sum(wi))
                self.weights[i][:] = (1-self.beta) * self.weights[i] + self.beta * wi

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'NTK'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


@keras_export('keras.callbacks.AdaptiveSampleWeight2')
class AdaptiveSampleWeight2(Callback):
    """ Callback that evaluate the adaptive weights based on the Gradient Pathologies approach by Wang et al.
    """
    def __init__(self, model, data_generator, loss=None,
                 beta=0.1, freq=100, log_freq=None,
                 hessian=False, types=None, **kwargs):
        super(AdaptiveSampleWeight2, self).__init__()
        # limit number of samples for performance concerns.
        self.data_generator = data_generator
        self.inputs = data_generator.get_grid()
        # eval loss and gradients.
        if loss is None:
            self.loss = K.function(model.inputs, model.outputs[0])
        else:
            self.loss = K.function(model.inputs, loss.outputs)
        self.freq = np.Inf if (freq==False or freq==0) else freq
        self.beta = beta
        if log_freq is None:
            self.log_freq = self.freq
        else:
            self.log_freq = log_freq

    def on_train_begin(self, logs={}):
        pass

    def on_epoch_begin(self, epoch, logs={}):
        if (epoch + 1) % self.freq == 0:
            loss = unpack_singleton(self.loss(self.inputs))
            self.data_generator.set_sample_weights(loss)

    @staticmethod
    def prepare_inputs(*args, **kwargs):
        kwargs['method'] = 'AdaptiveSampleWeight2'
        if len(args) == 1:
            kwargs['freq'] = args[0]
        elif len(args) > 0:
            raise ValueError
        return kwargs


def setup_adaptive_weight_callback(adaptive_weights, model, constraints, data_generator):
    """ Prepare adaptive_weight callback.

    # Arguments
        adaptive_weights:
        model:
        constraints:
        data_generator:
    :return:
    """
    if not isinstance(adaptive_weights, dict):
        adaptive_weights = NTKLossWeight.prepare_inputs(adaptive_weights)

    if adaptive_weights["method"].lower() in ("gp", "gradientpathology", "gradient_pathology"):
        return GradientPathologyLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].lower() in ("gn", "gradnorm", "grad_norm"):
        return GradNormLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].lower() in ("ntk", "neuraltangentkernel", "neural_tangent_kernel"):
        return NTKLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].lower() in ("id", "inversedirichlet", "inverse_dirichlet"):
        return InverseDirichletLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].lower() in ("sasw", "selfadaptivesampleweight", "self_adaptive_sample_weight"):
        return SelfAdaptiveSampleWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )
    
    elif adaptive_weights["method"].lower() in ("sa", "selfadaptive", "self_adaptive"):
        return SelfAdaptiveLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].lower() in ("clw", "curriculum", "curriculum_loss_weight"):
        return CurriculumLossWeight(
            model, data_generator=data_generator,
            types=[type(v).__name__ for v in constraints],
            **adaptive_weights
        )

    elif adaptive_weights["method"].upper() == "MTL":
        return MTLLossWeight(
            model, data_generator=data_generator,
            **adaptive_weights
        )

    elif adaptive_weights["method"].upper() == "SCORE":
        return ScoreLossWeight(
            model, data_generator=data_generator,
            **adaptive_weights
        )

    else:
        raise ValueError(
            'adaptive_weights method should be either of the followings: (GP, NTK)'
        )




import pytest
import sciann as sn
import json
import os
import shutil
from tensorflow.keras import optimizers as tf_optimizers
import numpy as np 


@pytest.fixture(scope="module")
def variable_x():
    return sn.Variable('x')


@pytest.fixture(scope="module")
def variable_y():
    return sn.Variable('y')


@pytest.fixture(scope="module")
def functional_fx(variable_x):
    return sn.Functional('fx', variable_x, 2*[10], 'tanh')


@pytest.fixture(scope="module")
def functional_gx(variable_x):
    return sn.Functional('gx', variable_x, 2*[10], 'tanh')


@pytest.fixture(scope="module")
def functional_fxy(variable_x, variable_y):
    return sn.Functional('fxy', [variable_x, variable_y], 2*[10], 'tanh')


@pytest.fixture(scope="module")
def functional_gxy(variable_x, variable_y):
    return sn.Functional('gxy', [variable_x, variable_y], 2*[10], 'tanh')


@pytest.fixture(scope="module")
def functional_hxy(variable_x, variable_y):
    return sn.Functional('hxy', [variable_x, variable_y], 2*[10], 'tanh', 
                         kernel_initializer=1., bias_initializer=0.)


@pytest.fixture(scope="module")
def functional_hxy_diffs(variable_x, variable_y, functional_hxy):
    h_x = sn.diff(functional_hxy, variable_x)
    h_y = sn.diff(functional_hxy, variable_y)
    h_xx = sn.diff(h_x, variable_x)
    h_yy = sn.diff(h_y, variable_y)
    h_xy = sn.diff(h_x, variable_y)
    return h_x, h_y, h_xx, h_yy, h_xy


@pytest.fixture(scope="module")
def model_fx(variable_x, functional_fx):
    return sn.SciModel(variable_x, functional_fx)


@pytest.fixture(scope="module")
def model_fx_gx(variable_x, functional_fx, functional_gx):
    return sn.SciModel(variable_x, [functional_fx, functional_gx])


@pytest.fixture(scope="module")
def train_data_fx():
    x = np.linspace(-1, 1, 101)
    y = np.tanh(x)
    return x, y


@pytest.fixture(scope="module")
def train_data_fx_gx():
    x = np.linspace(-1, 1, 101)
    y1 = np.tanh(x)
    y2 = np.tanh(x)
    return x, [y1, y2]


@pytest.fixture(scope="module")
def test_data_x():
    x = np.linspace(-1, 1, 101)
    return x


@pytest.fixture(scope="module")
def test_data_xy():
    x = np.linspace(-1, 1, 11)
    y = np.linspace(0, 2, 11)
    return list(np.meshgrid(x, y))


@pytest.fixture(scope="module")
def test_data_xy_dict(test_data_xy):
    return {'x': test_data_xy[0], 'y': test_data_xy[1]}


@pytest.fixture(scope="module")
def test_data_xy_dict_exception(test_data_xy):
    return {'xrand': test_data_xy[0], 'y': test_data_xy[1]}


@pytest.fixture(scope="module")
def expected_hxy(test_data_xy):
    return 10*np.tanh(10*np.tanh(test_data_xy[0] + test_data_xy[1]))


@pytest.fixture(scope="module")
def expected_diff_hxy(test_data_xy):
    x, y = test_data_xy
    h_x = 10*(1 - np.tanh(10*np.tanh(x + y))**2)*(10 - 10*np.tanh(x + y)**2)
    h_y = 10*(1 - np.tanh(10*np.tanh(x + y))**2)*(10 - 10*np.tanh(x + y)**2)
    h_xx = -20*(1 - np.tanh(10*np.tanh(x + y))**2)*(10 - 10*np.tanh(x + y)**2)**2*np.tanh(10*np.tanh(x + y)) - 10*(2 - 2*np.tanh(x + y)**2)*(10 - 10*np.tanh(10*np.tanh(x + y))**2)*np.tanh(x + y)
    h_yy = -20*(1 - np.tanh(10*np.tanh(x + y))**2)*(10 - 10*np.tanh(x + y)**2)**2*np.tanh(10*np.tanh(x + y)) - 10*(2 - 2*np.tanh(x + y)**2)*(10 - 10*np.tanh(10*np.tanh(x + y))**2)*np.tanh(x + y)
    h_xy = -20*(1 - np.tanh(10*np.tanh(x + y))**2)*(10 - 10*np.tanh(x + y)**2)**2*np.tanh(10*np.tanh(x + y)) - 10*(2 - 2*np.tanh(x + y)**2)*(10 - 10*np.tanh(10*np.tanh(x + y))**2)*np.tanh(x + y)
    return [h_x, h_y, h_xx, h_xy, h_yy]


def test_variable():
    xt = sn.Variable('xt', 10, dtype='float32')
    assert sn.is_variable(xt)
    assert sn.is_functional(xt)


def test_variable_exception():
    with pytest.raises(TypeError):
        x = sn.Variable(10)


def test_functional(variable_x):
    x = variable_x
    ft = sn.Functional('ft', x, 2*[10], 'tanh')
    assert sn.is_functional(ft)


def test_functional_exceptions(variable_x):
    x = variable_x
    with pytest.raises(TypeError):
        f = sn.Functional(x)
    with pytest.raises(TypeError):
        ft = sn.Functional('ft', 2*[10])
    with pytest.raises(TypeError):
        ft = sn.Functional('ft', x, 'tanh')
    with pytest.raises(TypeError):
        ft = sn.Functional('ft', x, 2*[10], 12)


def test_activation_function_exceptions(variable_x):
    x = variable_x
    assert sn.is_functional(
        sn.Functional('ft', x, 2*[10], 'l-tanh')
    )
    assert sn.is_functional(
        sn.Functional('gt', x, 2*[10], 'g-tanh')
    )
    assert sn.is_functional(
        sn.Functional('rt', x, 2*[10], ['tanh', 'sin', 'l-tanh'])
    )


def test_variable_operators(variable_x, variable_y):
    assert sn.is_functional(variable_x**2)
    assert sn.is_functional(sn.tanh(variable_x))
    assert sn.is_functional(variable_x + variable_y)
    assert sn.is_functional(sn.tanh(variable_x) + variable_y**2)
    assert sn.is_functional(variable_x + sn.sin(variable_y))
    assert sn.is_functional(2*variable_x + 5)
    assert sn.is_functional(variable_x*2 + 5)


def test_variable_operator_exceptions(variable_x, variable_y):
    with pytest.raises(AttributeError):
        a = variable_x + "to_fail"
    with pytest.raises(AttributeError):
        b = variable_y + None
    with pytest.raises(TypeError):
        c = variable_y**None
    with pytest.raises(TypeError):
        d = variable_y**[1,2,3]


def test_functional_operations(functional_fx, functional_gx, functional_fxy):
    assert sn.is_functional(functional_fx + functional_gx)
    assert sn.is_functional(functional_gx + functional_fxy)
    assert sn.is_functional(functional_fx * functional_gx)
    assert sn.is_functional(functional_fx + functional_gx*functional_fxy)
    assert sn.is_functional(functional_fx / functional_gx)


def test_functional_operation_exceptions(functional_fx, functional_gx, functional_fxy):
    with pytest.raises(TypeError):
        a = functional_fx ** functional_gx


def test_diff(variable_x, variable_y, functional_hxy):
    hxy_x = sn.diff(functional_hxy, variable_x)
    hxy_y = sn.diff(functional_hxy, variable_x)


def test_scimodel(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    assert isinstance(sn.SciModel(xs, ys), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mse", "adam"), sn.SciModel)
    assert isinstance(sn.SciModel(variable_x, ys), sn.SciModel)
    assert isinstance(sn.SciModel(xs, functional_fx), sn.SciModel)


def test_scimodel_exceptions(variable_x, variable_y, functional_fx, functional_gx):
    with pytest.raises(ValueError):
        sn.SciModel(variable_x)
    with pytest.raises(ValueError):
        sn.SciModel(functional_fx)


def test_scimodel_optimizers(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    assert isinstance(sn.SciModel(xs, ys, "mse", "adam"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mse", "rmsprop"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mse", "scipy-l-bfgs-b"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mse", "scipy-newtoncg"), sn.SciModel)


def test_scimodel_optimizer_exceptions(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    with pytest.raises(ValueError):
        assert isinstance(sn.SciModel(xs, ys, "mse", "to_fail"), sn.SciModel)


def test_scimodel_keras_optimizers(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    assert isinstance(sn.SciModel(xs, ys, "mse", tf_optimizers.Adam()), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mse", tf_optimizers.RMSprop()), sn.SciModel)


def test_scimodel_lossfuncs(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    assert isinstance(sn.SciModel(xs, ys, "mse", "adam"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "mae", "adam"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "ae", "adam"), sn.SciModel)
    assert isinstance(sn.SciModel(xs, ys, "sse", "adam"), sn.SciModel)


def test_scimodel_lossfunc_exceptions(variable_x, variable_y, functional_fx, functional_gx):
    xs = [variable_x, variable_y]
    ys = [functional_fx, functional_gx]
    with pytest.raises(ValueError):
        assert isinstance(sn.SciModel(xs, ys, "to_fail", "adam"), sn.SciModel)


def test_eval_functional(expected_hxy, test_data_xy, functional_hxy):
    hxy_test = functional_hxy.eval(test_data_xy)
    assert np.linalg.norm(hxy_test - expected_hxy) < 1e-6*np.linalg.norm(expected_hxy), \
        'test_eval_functional: failed using list inputs. '


def test_eval_functional_from_dict(expected_hxy, test_data_xy_dict, functional_hxy):
    hxy_test = functional_hxy.eval(test_data_xy_dict)
    assert np.linalg.norm(hxy_test - expected_hxy) < 1e-6 * np.linalg.norm(expected_hxy), \
        'test_eval_functional: failed using dict inputs. '


def test_eval_functional_from_dict_exception(expected_hxy, test_data_xy_dict_exception, functional_hxy):
    with pytest.raises(ValueError):
        functional_hxy.eval(test_data_xy_dict_exception)


def test_eval_diff(expected_diff_hxy, test_data_xy, functional_hxy_diffs):
    diff_test = [f.eval(test_data_xy) for f in functional_hxy_diffs]
    diff_error = [np.linalg.norm(test - true) < 1e-5*np.linalg.norm(true) 
                  for test, true in zip(diff_test, expected_diff_hxy)]
    assert all(diff_error)


def test_eval_scimodel(test_data_xy, expected_hxy, variable_x, variable_y, functional_hxy):
    m = sn.SciModel([variable_x, variable_y], functional_hxy)
    hxy_test = functional_hxy.eval(m, test_data_xy)
    assert np.linalg.norm(hxy_test - expected_hxy) < 1e-6*np.linalg.norm(expected_hxy)


def test_train(train_data_fx, model_fx):
    h = model_fx.train(train_data_fx[0], train_data_fx[1], epochs=10)


def test_train_learning_rate(train_data_fx, model_fx):
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10, 
        learning_rate = 0.001
    )


def test_train_learning_rate_expscheduler(train_data_fx, model_fx):
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10,
        learning_rate = ([0, 100], [0.001, 0.0001])
    )


def test_train_learning_rate_expscheduler(train_data_fx, model_fx):
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10, 
        learning_rate = {"scheduler": "exponentialdecay", 
                         "initial_learning_rate": 0.001, 
                         "final_learning_rate":0.0001,
                         "decay_epochs": 100}
    )


def test_train_learning_rate_sinescheduler(train_data_fx, model_fx):
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10, 
        learning_rate = {"scheduler": "sineexponentialdecay", 
                         "initial_learning_rate": 0.001, 
                         "final_learning_rate":0.0001,
                         "decay_epochs": 100,
                         "sine_freq": 2,
                         "sine_decay_rate": 0.5}
    )


def test_train_save_weights(train_data_fx, model_fx):
    name_prefix = "weights"
    default_path = os.path.join(os.curdir, name_prefix)
    # test 1: best weights 
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10,
        save_weights={"path":default_path, "freq": 10, "best": True}
    )
    files = list(filter(lambda f: f.startswith(name_prefix + "-best"), os.listdir(os.curdir)))
    assert len(files) > 0
    # test 2: save weights with freq
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10,
        save_weights={"path":default_path, "freq": 10, "best": False}
    )
    files = list(filter(lambda f: f.startswith(name_prefix), os.listdir(os.curdir)))
    assert len(files) > 1
    # delete temporary files. 
    for f in files:
        os.remove(os.path.join(os.curdir, f))


def test_train_log_functionals(train_data_fx, test_data_x, model_fx, functional_fx):
    dir_prefix = "logs"
    default_path = os.path.join(os.curdir, dir_prefix)
    # test 1: best weights 
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10,
        log_functionals={"functionals":[functional_fx], "path": default_path, 
                         "inputs":test_data_x, "freq": 5}
    )
    assert os.path.isdir(default_path)
    files = list(filter(lambda f: f.startswith("functional-history"), os.listdir(default_path)))
    assert len(files) > 1
    shutil.rmtree(default_path)


def test_train_log_loss_landscape(train_data_fx, test_data_x, model_fx, functional_fx):
    dir_prefix = "logs"
    default_path = os.path.join(os.curdir, dir_prefix)
    # test 1: best weights 
    h = model_fx.train(
        train_data_fx[0], train_data_fx[1], epochs=10,
        log_loss_landscape={"norm":2, "resolution":3, "path": default_path, "trials": 3}
    )
    assert os.path.isdir(default_path)
    files = list(filter(lambda f: f.startswith("loss-landscape"), os.listdir(default_path)))
    assert len(files) > 1
    shutil.rmtree(default_path)


def test_train_adaptive_weights(train_data_fx_gx, test_data_x, model_fx_gx):
    # test 1
    h = model_fx_gx.train(
        train_data_fx_gx[0], train_data_fx_gx[1], epochs=10,
        adaptive_weights=True
    )


def test_train_adaptive_weights_gradient_pathology(train_data_fx_gx, test_data_x, model_fx_gx):
    h = model_fx_gx.train(
        train_data_fx_gx[0], train_data_fx_gx[1], epochs=10,
        adaptive_weights={"method": "GP", "freq": 5, "alpha": 1.}
    )


def test_train_adaptive_weights_grad_norm(train_data_fx_gx, test_data_x, model_fx_gx):
    h = model_fx_gx.train(
        train_data_fx_gx[0], train_data_fx_gx[1], epochs=10,
        adaptive_weights={"method": "GN", "freq": 5, "alpha": 1.}
    )


def test_train_adaptive_weights_neural_tangent_kernel(train_data_fx_gx, test_data_x, model_fx_gx):
    h = model_fx_gx.train(
        train_data_fx_gx[0], train_data_fx_gx[1], epochs=10,
        adaptive_weights={"method": "NTK", "freq": 5, "alpha": 1.}
    )


if __name__ == '__main__':
    pytest.main(['--verbose'])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras import backend as K
graph_unique_name = K.get_graph().unique_name

from keras.layers import InputLayer

from ..utils import to_list
from ..utils import Lambda

from .functional import MLPFunctional
from .variable import Variable


class RadialBasis(MLPFunctional):
    """ Radial Basis functional class.
    """
    def __init__(self, vars, rb_vars, radii):
        vars = to_list(vars)
        if not all([isinstance(x, Variable) for x in vars]):
            raise TypeError

        rb_vars = to_list(rb_vars)
        if not all([isinstance(x, RadialBasisBase) for x in rb_vars]):
            try:
                for i, rbv in enumerate(rb_vars):
                    rb_vars[i] = RadialBasisBase(rbv)
            except (ValueError, TypeError):
                raise ValueError('Expected `str` or `RadialBasisBase` as rb_vars. ')

        if len(vars) != len(rb_vars):
            raise ValueError

        if radii <= 0.0:
            raise ValueError('Expecting a positive value for `radii`. ')

        inputs, layers = [], []
        for v in vars:
            inputs += v.outputs
            layers += v.layers
        for v in rb_vars:
            inputs += v.outputs
            layers += v.layers

        outputs = []
        for i in range(len(vars)):
            l = Lambda(
                lambda x: K.exp(-(x[1] - x[0])**2/radii**2),
                name = graph_unique_name("{}/".format('RadialBasis'))
            )
            assert len(vars[i].outputs) == 1
            assert len(rb_vars[i].outputs) == 1
            layers.append(l)
            outputs.append(l(vars[i].outputs + rb_vars[i].outputs))

        super(RadialBasis, self).__init__(
            layers=layers,
            inputs=inputs,
            outputs=outputs
        )

    @classmethod
    def get_class(cls):
        return MLPFunctional


class RadialBasisBase(MLPFunctional):
    """ Configures the `RadialBasisBase` object for the network's input.

    # Arguments
        name: String.
            Required as derivatives work only with layer names.
        units (Int): Number of nodes to the network.
            Minimum number is 1.
        tensor: Tensorflow `Tensor`.
            Can be pass as the input path.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').

    # Raises
        ValueError: Provide `units > 0`.
    """
    def __init__(self,
                 name=None,
                 units=1,
                 tensor=None,
                 dtype=None):

        if not dtype:
            dtype = K.floatx()
        elif not dtype == K.floatx():
            K.set_floatx(dtype)

        if units < 1:
            raise ValueError(
                'Expected at least one unit size - was provided `units`={:d}'.format(units)
            )

        layer = InputLayer(
            batch_input_shape=(None, units),
            input_tensor=tensor,
            name=name,
            dtype=dtype
        )

        super(RadialBasisBase, self).__init__(
            layers=to_list(layer),
            inputs=to_list(layer.input),
            outputs=to_list(layer.output),
        )

    @classmethod
    def get_class(cls):
        return MLPFunctional




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from . import field
from . import functional
from . import parameter
from . import radialbasis
from . import variable

from . import rnn_field
from . import rnn_functional
from . import rnn_variable

from .functional import Functional
from .field import Field
from .parameter import Parameter
from .variable import Variable
# from .fourier import FourierLayer
from .radialbasis import RadialBasis

from .rnn_functional import RNNFunctional
from .rnn_variable import RNNVariable
from .rnn_field import RNNField



""" SciModel class to define and train the model.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import tensorflow as tf
from keras import backend as K
from tensorflow.python import keras as k
import numpy as np

from keras.utils.data_utils import Sequence
from keras.models import Model
from keras.utils.vis_utils import plot_model
from tensorflow import gradients as tf_gradients

from ..utils import unpack_singleton, to_list
from ..utils import is_variable, is_constraint, is_functional
from ..utils.optimizers import GradientObserver, ScipyOptimizer, GeneratorWrapper
from ..utils.callbacks import *
from ..utils.schedules import setup_lr_scheduler

from ..functionals import Variable
from ..functionals import RadialBasis
from ..constraints import Data, PDE, Tie


class SciModel(object):
    """Configures the model for training.
    Example:
    # Arguments
        inputs: Main variables (also called inputs, or independent variables) of the network, `xs`.
            They all should be of type `Variable`.
        targets: list all targets (also called outputs, or dependent variables)
            to be satisfied during the training. Expected list members are:
            - Entries of type `Constraint`, such as Data, Tie, etc.
            - Entries of type `Functional` can be:
                . A single `Functional`: will be treated as a Data constraint.
                    The object can be just a `Functional` or any derivatives of `Functional`s.
                    An example is a PDE that is supposed to be zero.
                . A tuple of (`Functional`, `Functional`): will be treated as a `Constraint` of type `Tie`.
            - If you need to impose more complex types of constraints or
                to impose a constraint partially in a specific part of region,
                use `Data` or `Tie` classes from `Constraint`.
        loss_func: defaulted to "mse" or "mean_squared_error".
            It can be an string from supported loss functions, i.e. ("mse" or "mae").
            Alternatively, you can create your own loss function and
            pass the function handle (check Keras for more information).
        optimizer: defaulted to "adam" optimizer.
            It can be one of Keras accepted optimizers, e.g. "adam".
            You can also pass more details on the optimizer:
            - `optimizer = k.optimizers.RMSprop(lr=0.01, rho=0.9, epsilon=None, decay=0.0)`
            - `optimizer = k.optimizers.SGD(lr=0.001, momentum=0.0, decay=0.0, nesterov=False)`
            - `optimizer = k.optimizers.Adam(lr=0.01, beta_1=0.9, beta_2=0.999, epsilon=None, decay=0.0, amsgrad=False)`
            Check our Keras documentation for further details. We have found
        load_weights_from: (file_path) Instantiate state of the model from a previously saved state.
        plot_to_file: A string file name to output the network architecture.

    # Raises
        ValueError: `inputs` must be of type Variable.
                    `targets` must be of types `Functional`, or (`Functional`, data), or (`Functional`, `Functional`).
    """
    def __init__(self,
                 inputs=None,
                 targets=None,
                 loss_func="mse",
                 optimizer="adam",
                 load_weights_from=None,
                 plot_to_file=None,
                 **kwargs):
        # strictly check for inputs to be of type variable.
        inputs = to_list(inputs)
        if not all([is_variable(x) for x in inputs]):
            raise ValueError(
                'Please provide a `list` of `Variable` or `RadialBasis` objects for inputs. '
            )
        # prepare input tensors.
        input_vars = []
        for var in inputs:
            input_vars += var.inputs
        # check outputs if of correct type.
        if targets is None:
            if 'constraints' in kwargs:
                targets = kwargs.get('constraints')
            elif 'conditions' in kwargs:
                targets = kwargs.get('conditions')
        else:
            if 'conditions' in kwargs or 'constraints' in kwargs:
                raise TypeError(
                    'Inconsistent inputs: `constraints`, `conditions`, and `targets` are all equivalent keywords '
                    '- pass all targets as a list to `SciModel`. '
                )
        # setup constraints.
        targets = to_list(targets)
        for i, y in enumerate(targets):
            if not is_constraint(y):
                if is_functional(y):
                    # Case of Data-type constraint.
                    # By default, targets are initialized with Data.
                    targets[i] = Data(y)
                elif isinstance(y, tuple) and \
                        len(y) == 2 and \
                        is_functional(y[0]) and is_functional(y[1]):
                    # Case of Tie-type constraint.
                    targets[i] = Tie(y[0], y[1])
                else:
                    # Not recognised.
                    raise ValueError(
                        'The {}th target entry is not of type `Constraint` or `Functional` - '
                        'received \n ++++++ {} '.format(i, y)
                    )
        # prepare network outputs.
        output_vars = []
        for cond in targets:
            output_vars += cond().outputs
        # prepare loss_functions.
        if not isinstance(loss_func, list):
            loss_func = len(output_vars)*[loss_func]
        # prepare the losses.
        for i, li in enumerate(loss_func):
            if isinstance(li, str):
                loss_func[i] = SciModel.loss_functions(li)
            elif not callable(li):
                raise TypeError(
                    'Please provide a valid loss function from ("mse", "mae") '
                    + "or a callable function for input of tensor types. "
                )
        # Initialize the Model form super class.
        model = Model(
            inputs=input_vars,
            outputs=output_vars,
            **kwargs
        )

        # compile the model.
        loss_weights = [K.variable(1.0) for v in output_vars]

        # Set the variables.
        self._model = model
        self._inputs = inputs
        self._constraints = targets
        self._loss_func = loss_func
        self._optimizer = optimizer
        self._loss_weights = loss_weights
        self._callbacks = {}
        
        # compile the model 
        self.compile()
        
        # set initial state of the model.
        if load_weights_from is not None:
            if os.path.exists(load_weights_from): 
                model.load_weights(load_weights_from)
            else:
                raise Warning("File not found - load_weights_from: {}".format(load_weights_from))

        # Plot to file if requested.
        if plot_to_file is not None:
            plot_model(self._model, to_file=plot_to_file)

    @property
    def model(self):
        return self._model

    @property
    def constraints(self):
        return self._constraints

    @property
    def inputs(self):
        return self._inputs

    def compile(self):
        loss_func = self._loss_func 
        loss_weights = self._loss_weights
        optimizer = self._optimizer

        if isinstance(optimizer, str) and \
                len(optimizer.lower().split("scipy-")) > 1:
            optimizer = GradientObserver(method=optimizer)

        self._model.compile(loss=loss_func,
                            optimizer=optimizer,
                            loss_weights=loss_weights)

    def clear_callbacks(self):
        self._callbacks.clear()

    def load_weights(self, file):
        if os.path.exists(file):
            self.model.load_weights(file)
        else:
            raise ValueError('File not found.')

    def verify_update_constraints(self, constraints):
        ver = []
        for old, new in zip(self._constraints, constraints):
            if old==new and old.sol==new.sol:
                if old.sol is None:
                    ver.append(True)
                else:
                    if all([all(xo==xn) for xo, xn in zip(old.sol, new.sol)]):
                        ver.append(True)
                    else:
                        ver.append(False)
            else:
                ver.append(False)
        return all(ver)

    def __call__(self, *args, **kwargs):
        output = self._model.__call__(*args, **kwargs)
        return output if isinstance(output, list) else [output]

    def save(self, filepath, *args, **kwargs):
        return self._model.save(filepath, *args, **kwargs)

    def save_weights(self, filepath, *args, **kwargs):
        return self._model.save_weights(filepath, *args, **kwargs)

    def load_weights(self, filepath, *args, **kwargs):
        return self._model.load_weights(filepath, *args, **kwargs)

    def summary(self, *args, **kwargs):
        return self._model.summary(*args, **kwargs)

    def train(self,
              x_true,
              y_true=None,
              weights=None,
              target_weights=None,
              batch_size=2**6,
              epochs=100,
              learning_rate=0.001,
              adaptive_weights=None,
              adaptive_sample_weights=None,
              log_loss_gradients=None,
              shuffle=True,
              callbacks=[],
              stop_lr_value=1e-8,
              reduce_lr_after=None,
              reduce_lr_min_delta=0.,
              stop_after=None,
              stop_loss_value=1e-8,
              log_parameters=None,
              log_functionals=None,
              log_loss_landscape=None,
              save_weights=None,
              default_zero_weight=0.0,
              validation_data=None,
              **kwargs):
        """Performs the training on the model.

        # Arguments
            x_true: list of `Xs` associated to targets of `Y`.
                Alternatively, you can pass a Sequence object (keras.utils.Sequence).
                Expecting a list of np.ndarray of size (N,1) each,
                with N as the sample size.
            y_true: list of true `Ys` associated to the targets defined during model setup.
                Expecting the same size as list of targets defined in `SciModel`.
                - To impose the targets at specific `Xs` only, pass a tuple of `(ids, y_true)` for that target.
            weights: (np.ndarray) A global sample weight to be applied to samples.
                Expecting an array of shape (N,1), with N as the sample size.
                Default value is `one` to consider all samples equally important.
            target_weights: (list) A weight for each target defined in `y_true`.
            batch_size: (Integer) or 'None'.
                Number of samples per gradient update.
                If unspecified, 'batch_size' will default to 2^6=64.
            epochs: (Integer) Number of epochs to train the model.
                Defaulted to 100.
                An epoch is an iteration over the entire `x` and `y`
                data provided.
            learning_rate: (Tuple/List) (epochs, lrs).
                Expects a list/tuple with a list of epochs and a list or learning rates.
                It linearly interpolates between entries.
                Defaulted to 0.001 with no decay.
                Example:
                    learning_rate = ([0, 100, 1000], [0.001, 0.0005, 0.00001])
            shuffle: Boolean (whether to shuffle the training data).
                Default value is True.
            adaptive_weights: Pass a Dict with the following keys:
                . method: GP or NTK.
                . freq: Freq to update the weights.
                . log_freq: Freq to log the weights and gradients in the history object.
                . beta: The beta parameter in from Gradient Pathology paper.
            log_loss_gradients: Pass a Dict with the following keys:
                . freq: Freq of logs. Defaulted to 100.
                . path: Path to log the gradients.
            callbacks: List of `keras.callbacks.Callback` instances.
            reduce_lr_after: patience to reduce learning rate or stop after certain missed epochs.
                Defaulted to epochs max(10, epochs/10).
            stop_lr_value: stop the training if learning rate goes lower than this value.
                Defaulted to 1e-8.
            reduce_lr_min_delta: min absolute change in total loss value that is considered a successful change.
                Defaulted to 0.001. 
                This values affects number of failed attempts to trigger reduce learning rate based on reduce_lr_after. 
            stop_after: To stop after certain missed epochs. Defaulted to total number of epochs.
            stop_loss_value: The minimum value of the total loss that stops the training automatically. 
                Defaulted to 1e-8.
            log_parameters: Dict object expecting the following keys:
                . parameters: pass list of parameters.
                . freq: pass freq of outputs.
            log_functionals: Dict object expecting the following keys:
                . functionals: List of functionals to log their training history.
                . inputs: The input grid to evaluate the value of each functional.
                          Should be of the same size as the inputs to the model.train.
                . path: Path to the location that the csv files will be logged.
                . freq: Freq of logging the functionals.
            log_loss_landscape: Dict object expecting the following arguments:
                . norm: defaulted to 2.
                . resolution: defaulted to 10.
                . path: Path to the location that the csv files will be logged.
                . trials: Number of trials to pick a path.
            save_weights: (dict) Dict object expecting the following information:
                . path: defaulted to the current path.
                . freq: freq of calling CheckPoint callback.
                . best: If True, only saves the best loss, otherwise save all weights at every `freq` epochs.
            default_zero_weight: a small number for zero sample-weight.

        # Returns
            A Keras 'History' object after performing fitting.
        """
        if isinstance(x_true, Sequence):
            if y_true is not None:
                raise ValueError(
                    'If data_generator is provided, you should not provide y_true.'
                )
            data_generator = x_true

        else:
            if y_true is None:
                raise ValueError('(X_true, Y_true): Please provide proper values for both inputs and targets of SciModel. ')
            # prepare X,Y data.
            x_true = to_list(x_true)
            for i, (x, xt) in enumerate(zip(x_true, self._model.inputs)):
                x_shape = tuple(xt.get_shape().as_list())
                if x.shape != x_shape:
                    try:
                        x_true[i] = x.reshape((-1,) + x_shape[1:])
                    except:
                        print(
                            'Could not automatically convert the inputs to be ' 
                            'of the same size as the expected input tensors. ' 
                            'Please provide inputs of the same dimension as the `Variables`. '
                        )
                        assert False

            y_true = to_list(y_true)
            assert len(y_true)==len(self._constraints), \
                'Miss-match between expected targets (constraints) defined in `SciModel` and ' \
                'the provided `y_true`s - expecting the same number of data points. '

            num_sample = x_true[0].shape[0]
            assert all([x.shape[0]==num_sample for x in x_true[1:]]), \
                'Inconsistent sample size among `Xs`. '
            ids_all = np.arange(0, num_sample)

            if weights is None:
                weights = np.ones(num_sample)
            elif isinstance(weights, np.ndarray):
                if len(weights.shape)!=1 or \
                        weights.shape[0] != num_sample:
                    try:
                        weights = weights.reshape(num_sample)
                    except:
                        raise ValueError(
                            'Input error: `weights` should have dimension 1 with '
                            'the same sample length as `Xs. '
                        )

            sample_weights, y_star = [], []
            for i, yt in enumerate(y_true):
                c = self._constraints[i]
                # verify entry.
                ys, wei = SciModel._prepare_data(
                    c.cond.outputs, to_list(yt),
                    weights if isinstance(weights, np.ndarray) else weights[i],
                    num_sample, default_zero_weight
                )
                # add to the list.
                y_star += ys
                sample_weights += wei

            # create Sequence wrapper for training. 
            data_generator = GeneratorWrapper(
                x_true, y_star, sample_weights, 
                batch_size, shuffle
            )

        # initialize callbacks.
        sci_callbacks = []
        # Learning rate setup.
        K.set_value(self.model.optimizer.lr, 1e-3)
        if reduce_lr_after is None:
            reduce_lr_after = max([10, epochs / 10])
        if stop_after is None:
            stop_after = epochs
        if isinstance(learning_rate, (type(None), float, int)):
            lr_rates = 0.001 if learning_rate is None else learning_rate
            K.set_value(self.model.optimizer.lr, lr_rates)
            lr_scheduler = {
                "scheduler": "default",
                "reduce_lr_after": reduce_lr_after,
                "reduce_lr_min_delta": reduce_lr_min_delta
            }
        elif isinstance(learning_rate, (tuple, list)):
            lr_scheduler = {
                "scheduler": "learning_rate_scheduler",
                "lr_epochs": learning_rate[0],
                "lr_values": learning_rate[1]
            }
        elif isinstance(learning_rate, dict):
            lr_scheduler = learning_rate
        else:
            raise ValueError(
                "learning rate: expecting a `float` or \n  "
                "a tuple/list of two arrays with `epochs` and `learning rates` or \n  "
                "a dictionary with scheduler settings. "
            )
        if 'lr_scheduler' in self._callbacks and learning_rate is None:
            sci_callbacks.append(self._callbacks['lr_scheduler'])
        else:
            sci_callbacks.append(setup_lr_scheduler(lr_scheduler))
            self._callbacks['lr_scheduler'] = sci_callbacks[-1]

        sci_callbacks += [
            k.callbacks.EarlyStopping(monitor="loss", mode='auto', verbose=1,
                                      patience=stop_after, min_delta=1e-9),
            k.callbacks.TerminateOnNaN(),
            EarlyStoppingByLossVal(stop_loss_value),
            EarlyStoppingByLearningRate(stop_lr_value),
            EpochTime()
        ]

        # setup target weights.
        len_inputs = len(to_list(self._model.inputs))
        len_outputs = len(to_list(self._model.outputs))
        if target_weights is not None:
            if not(isinstance(target_weights, list) and
                   len(target_weights) == len_outputs):
                raise ValueError(
                    'Expected a list of weights for the same size as the targets '
                    '- was provided {}'.format(target_weights)
                )
        else:
            target_weights = len_outputs * [1.0]

        # save model.
        model_file_path = None
        if save_weights is not None:
            assert isinstance(save_weights, dict), \
                "pass a dictionary containing `path, freq, best`. \n" \
                "e.g. save_weights = dict(path='output/path/', freq=10, best=False)"
            save_weights_path = save_weights.get('path', os.path.join(os.curdir, "weights"))
            try:
                if save_weights.get('best', False) == True:
                    model_file_path = save_weights_path + "-best.hdf5"
                    model_check_point = k.callbacks.ModelCheckpoint(
                        model_file_path, monitor='loss', save_weights_only=True, mode='auto',
                        period=save_weights.get('freq', 10),
                        save_best_only=True
                    )
                else:
                    self._model.save_weights("{}-start.hdf5".format(save_weights_path))
                    model_file_path = save_weights_path + "-{epoch:05d}-{loss:.3e}.hdf5"
                    model_check_point = k.callbacks.ModelCheckpoint(
                        model_file_path, monitor='loss', save_weights_only=True, mode='auto',
                        period=save_weights.get('freq', 10),
                        save_best_only=False
                    )
            except:
                print("\nWARNING: Failed to save model.weights to the provided path: {}\n".format(save_weights_path))

        if model_file_path is not None:
            sci_callbacks.append(model_check_point)

        if isinstance(self._model.optimizer, GradientObserver):
            opt = ScipyOptimizer(self._model)
            opt_fit_func = opt.fit
        else:
            opt_fit_func = self._model.fit

        if adaptive_weights:
            sci_callbacks.append(
                setup_adaptive_weight_callback(
                    adaptive_weights=adaptive_weights,
                    model=self.model,
                    constraints=self.constraints,
                    data_generator=data_generator)
            )
            self._callbacks['adaptive_weights'] = sci_callbacks[-1]
        elif 'adaptive_weights' in self._callbacks:
            sci_callbacks.append(self._callbacks['adaptive_weights'])

        if adaptive_sample_weights:
            # sample_weights = [K.variable(wi) for wi in sample_weights]
            sci_callbacks.append(
                AdaptiveSampleWeight2(
                    self.model, data_generator=data_generator,
                    types=[type(v).__name__ for v in self.constraints],
                    **adaptive_sample_weights
                )
            )
            # loss_gradients = NTKSW.eval_diag_ntk()
            # sample_weights = NTKSW.eval_sample_weights(loss_gradients)
            self._callbacks['adaptive_sample_weights'] = sci_callbacks[-1]

        elif 'adaptive_sample_weights' in self._callbacks:
            sci_callbacks.append(self._callbacks['adaptive_sample_weights'])

        if log_loss_gradients:
            if not isinstance(log_loss_gradients, dict):
                log_loss_gradients = LossGradientHistory.prepare_inputs(log_loss_gradients)
            sci_callbacks.append(
                LossGradientHistory(
                    self.model, data_generator=data_generator,
                    **log_loss_gradients
                )
            )
            self._callbacks['log_loss_gradients'] = sci_callbacks[-1]
        elif 'log_loss_gradients' in self._callbacks:
            sci_callbacks.append(self._callbacks['log_loss_gradients'])

        if log_parameters:
            if not isinstance(log_parameters, dict):
                log_parameters = ParameterHistory.prepare_inputs(log_parameters)
            sci_callbacks.append(
                ParameterHistory(**log_parameters)
            )
            self._callbacks['log_parameters'] = sci_callbacks[-1]
        elif 'log_parameters' in self._callbacks:
            sci_callbacks.append(self._callbacks['log_parameters'])

        if log_functionals:
            if not isinstance(log_functionals, dict):
                log_functionals = FunctionalHistory.prepare_inputs(log_functionals)
            sci_callbacks.append(
                FunctionalHistory(**log_functionals)
            )
            self._callbacks['log_functionals'] = sci_callbacks[-1]
        elif 'log_functionals' in self._callbacks:
            sci_callbacks.append(self._callbacks['log_functionals'])

        if log_loss_landscape:
            if not isinstance(log_loss_landscape, dict):
                log_loss_landscape = LossLandscapeHistory.prepare_inputs(log_loss_landscape)
            sci_callbacks.append(
                LossLandscapeHistory(
                    self.model, data_generator=data_generator,
                    **log_loss_landscape
                )
            )
            self._callbacks['log_loss_landscape'] = sci_callbacks[-1]
        elif 'log_loss_landscape' in self._callbacks:
            sci_callbacks.append(self._callbacks['log_loss_landscape'])

        # training the models.
        history = opt_fit_func(
            data_generator,  # sums to number of samples.
            epochs=epochs,
            callbacks=to_list(sci_callbacks) + to_list(callbacks),
            validation_data=validation_data,
            **kwargs
        )

        if save_weights is not None:
            if 'best' not in save_weights.keys() or \
                    save_weights['best'] is False:
                try:
                    self._model.save_weights("{}-end.hdf5".format(save_weights_path))
                except:
                    print("\nWARNING: Failed to save model.weights to the provided path: {}\n".format(save_weights_path))

        # return the history.
        return history

    def predict(self, xs,
                batch_size=None,
                verbose=0,
                steps=None):
        """ Predict output from network.

        # Arguments
            xs: list of `Xs` associated model.
                Expecting a list of np.ndarray of size (N,1) each,
                with N as the sample size.
            batch_size: defaulted to None.
                Check Keras documentation for more information.
            verbose: defaulted to 0 (None).
                Check Keras documentation for more information.
            steps: defaulted to 0 (None).
                Check Keras documentation for more information.

        # Returns
            List of numpy array of the size of network outputs.

        # Raises
            ValueError if number of `xs`s is different from number of `inputs`.
        """
        xs = to_list(xs)
        if len(xs) != len(self._inputs):
            raise ValueError(
                "Please provide consistent number of inputs as the model is defined: "
                "Expected {} - provided {}".format(len(self._inputs), len(to_list(xs)))
            )
        # To have unified output for postprocessing - limitted support.
        shape_default = [x.shape for x in xs]
        assert all([shape_default[0][0]==x[0] for x in shape_default[1:]])
        # prepare X,Y data.
        for i, (x, xt) in enumerate(zip(xs, self._model.inputs)):
            x_shape = tuple(xt.get_shape().as_list())
            if x.shape != x_shape:
                try:
                    xs[i] = x.reshape((-1,) + x_shape[1:])
                except:
                    print(
                        'Could not automatically convert the inputs to be ' 
                        'of the same size as the expected input tensors. ' 
                        'Please provide inputs of the same dimension as the `Variables`. '
                    )
                    assert False

        y_pred = to_list(self._model.predict(xs, batch_size, verbose, steps))

        # return uniform shapes. 
        if all([shape_default[0]==sd for sd in shape_default[1:]]):
            try:
                y_pred = [y.reshape(shape_default[0]) for y in y_pred]
            except:
                print("Input and output dimensions need re-adjustment for post-processing.")

        # revert back to normal.
        for i, sd in enumerate(shape_default):
            xs[i] = xs[i].reshape(sd)
        xs = unpack_singleton(xs)

        return unpack_singleton(y_pred)

    def eval(self, *args):
        if len(args) == 1:
            x_data = to_list(args[0])
            if len(x_data) != len(self._inputs):
                raise ValueError(
                    "Please provide consistent number of inputs as the model is defined: "
                    "Expected {} - provided {}".format(len(self._inputs), len(x_data))
                )
            if not all([isinstance(xi, np.ndarray) for xi in x_data]):
                raise ValueError("Please provide input data to the network. ")
            return unpack_singleton(self.predict(x_data))

        elif len(args) == 2:
            var_name = args[0]
            if not isinstance(var_name, str):
                raise ValueError("Value Error: Expected a LayerName as the input. ")
            x_data = to_list(args[1])
            new_model = K.function(self._model.inputs, self._model.get_layer(var_name).output)
            if not all([isinstance(xi, np.ndarray) for xi in x_data]):
                raise ValueError("Please provide input data to the network. ")
            return unpack_singleton(new_model(x_data))

    def plot_model(self, *args, **kwargs):
        """ Keras plot_model functionality.
            Refer to Keras documentation for help.
        """
        plot_model(self._model, *args, **kwargs)

    @staticmethod
    def loss_functions(method="mse"):
        """ loss_function returns the callable object to evaluate the loss.

        # Arguments
            method: String.
            - "mse" for `Mean Squared Error` or
            - "mae" for `Mean Absolute Error` or
            - "se" for `Squared Error` or
            - "ae" for `Absolute Error` or
            - "sse" for `Squared Sum Error` or
            - "ase" for `Absolute Sum Error` or
            - "rse" for `Reduce Sum Error`.

        # Returns
            Callable function that gets (y_true, y_pred) as the input and
                returns the loss value as the output.

        # Raises
            ValueError if anything other than "mse" or "mae" is passed.
        """
        if method in ("mse", "mean_squared_error"):
            return lambda y_true, y_pred: K.mean(K.square(y_true - y_pred), axis=-1)
        elif method in ("mae", "mean_absolute_error"):
            return lambda y_true, y_pred: K.mean(K.abs(y_true - y_pred), axis=-1)
        elif method in ("se", "squared_error"):
            return lambda y_true, y_pred: K.sum(K.square(y_true - y_pred), axis=-1)
        elif method in ("ae", "absolute_error"):
            return lambda y_true, y_pred: K.sum(K.abs(y_true - y_pred), axis=-1)
        elif method in ("sse", "squared_sum_error"):
            return lambda y_true, y_pred: K.sum(K.square(K.sum(y_true - y_pred, axis=0, keepdims=True)), axis=-1)
        elif method in ("ase", "absolute_sum_error"):
            return lambda y_true, y_pred: K.sum(K.abs(K.sum(y_true - y_pred, axis=0, keepdims=True)), axis=-1)
        elif method in ("rse", "reduce_sum_error"):
            return lambda y_true, y_pred: K.sum(K.sum(y_true - y_pred, axis=0, keepdims=True), axis=-1)
        elif hasattr(k.losses, method):
            return getattr(k.losses, method)
        else:
            raise ValueError(
                'Supported losses: Keras loss function or (mse, mae, se, ae, sse, ase, rse)'
            )

    @staticmethod
    def _prepare_data(cond_outputs, y_true, global_weights, num_sample, default_zero_weight):
        ys, weis = [], []
        ids_all = np.arange(0, num_sample)
        # prepare sample weight.
        for i, yt in enumerate(to_list(y_true)):
            ids = None
            yc = cond_outputs[i]
            if isinstance(yt, tuple) and len(yt) == 2:
                ids = yt[0].flatten()
                if isinstance(yt[1], np.ndarray):
                    adjusted_yt = yt[1]
                    if ids.size == yt[1].shape[0] and ids.size < num_sample:
                        adjusted_yt = np.zeros((num_sample,)+yt[1].shape[1:])
                        adjusted_yt[ids, :] = yt[1]
                    elif yt[1].shape[0] != num_sample:
                        raise ValueError(
                            'Error in size of the target {}.'.format(i)
                        )
                else:
                    adjusted_yt = yt[1]
                ys.append(adjusted_yt)
            elif isinstance(yt, (np.ndarray, str, float, int, type(None))):
                ys.append(yt)
            else:
                raise ValueError(
                    'Unrecognized entry - please provide a list of `data` or tuples of `(ids, data)`'
                    ' for each target defined in `SciModel`. '
                )
            # Define weights of samples.
            if ids is None:
                ids = ids_all
                wei = global_weights
            else:
                wei = np.zeros(num_sample) + default_zero_weight
                wei[ids] = global_weights[ids]
                wei[ids] *= sum(global_weights)/sum(wei[ids])
            weis.append(wei)
            # preparing targets.
            if isinstance(ys[-1], np.ndarray):
                if not (ys[-1].shape[1:] == k.backend.int_shape(yc)[1:]):
                    try:
                        ys[-1] = ys[-1].reshape((-1,) + k.backend.int_shape(yc)[1:])
                    except (ValueError, TypeError):
                        raise ValueError(
                            'Dimension of expected `y_true` does not match with defined `Constraint`'
                        )
            elif isinstance(ys[-1], str):
                if ys[-1] == 'zero' or ys[-1] == 'zeros':
                    ys[-1] = np.zeros((num_sample, ) + k.backend.int_shape(yc)[1:])
                elif ys[-1] == 'one' or ys[-1] == 'ones':
                    ys[-1] = np.ones((num_sample, ) + k.backend.int_shape(yc)[1:])
                else:
                    raise ValueError(
                        'Unexpected `str` entry - only accepts `zeros` or `ones`.'
                    )
            elif isinstance(ys[-1], (int, float)):
                ys[-1] = np.ones((num_sample, ) + k.backend.int_shape(yc)[1:]) * float(ys[-1])
            elif isinstance(ys[-1], type(None)):
                ys[-1] = np.zeros((num_sample, ) + k.backend.int_shape(yc)[1:])
            else:
                raise ValueError(
                    'Unsupported entry - {} '.format(ys[-1])
                )
            # set undefined ids to zeros.
            if ids.size != num_sample:
                adjusted_ids = np.ones(num_sample, dtype=bool)
                adjusted_ids[ids] = False
                ys[-1][adjusted_ids, :] = 0.0

        return ys, weis




""" Functional class for SciANN.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras import backend as K
graph_unique_name = K.get_graph().unique_name

from keras.layers import Dense
from keras.layers import Activation
from keras.layers import Concatenate
from keras.layers import Lambda
from keras.models import Model
from tensorflow import tensordot, expand_dims
import numpy as np

from ..utils import to_list, unpack_singleton, is_same_tensor, unique_tensors
from ..utils import default_weight_initializer
from ..utils import default_regularizer
from ..utils import validations, getitem
from ..utils import floatx, set_floatx
from ..utils import math
# from ..utils.transformers import FourierFeature
from ..utils.activations import SciActivation, get_activation
from ..utils import prepare_default_activations_and_initializers

from .field import Field


class MLPFunctional(object):
    """ Configures the Functional object (Neural Network).

    # Arguments
        fields: String or Field.
            [Sub-]Network outputs.
            It can be of type `String` - Associated fields will be created internally.
            It can be of type `Field` or `Functional`
        variables: Variable.
            [Sub-]Network inputs.
            It can be of type `Variable` or other Functional objects.
        hidden_layers: A list indicating neurons in the hidden layers.
            e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
        activation: defaulted to "tanh".
            Activation function for the hidden layers.
            Last layer will have a linear output.
        output_activation: defaulted to "linear".
            Activation function to be applied to the network output.
        res_net: (True, False). Constructs a resnet architecture.
            Defaulted to False.
        kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
        bias_initializer: Initializer of the `Bias`, from `k.initializers`.
        kernel_regularizer: Regularizer for the kernel.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        bias_regularizer: Regularizer for the bias.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').
            Note: Only network inputs should be set.
        trainable: Boolean.
            False if network is not trainable, True otherwise.
            Default value is True.

    # Raises
        ValueError:
        TypeError:
    """
    def __init__(self, inputs, outputs, layers):
        # check data-type.
        self.inputs = inputs.copy()
        self.outputs = outputs.copy()
        self.layers = layers.copy()
        self._set_model()

    def _set_model(self, inputs, outputs, layers):
        self.inputs = inputs
        self.outputs = outputs
        self.layers = layers

    def eval(self, *args):
        """ Evaluates the functional object for a given input.

        # Arguments
            (SciModel, Xs): 
                Evalutes the functional object from the beginning 
                    of the graph defined with SciModel. 
                    The Xs should match those of SciModel. 
            
            (Xs):
                Evaluates the functional object from inputs of the functional. 
                    Xs should match those of inputs to the functional.

            (Xs):
                A dictionary, containing values for each variable.
                    
        # Returns
            Numpy array of dimensions of network outputs. 

        # Raises
            ValueError:
            TypeError:
        """
        if len(args) == 1:
            model = self.model
            # read data.
            xs = args[0]
        elif len(args) == 2:
            if validations.is_scimodel(args[0]):
                model = K.function(args[0].model.inputs, self.outputs)
            else:
                raise ValueError(
                    'Expected a SciModel object for the first arg. '
                )
            xs = args[1]
        else:
            raise NotImplemented()

        # To have unified output for postprocessing - limitted support.
        if isinstance(xs, dict):
            xs_names = []
            xs_vals = []
            for x_in in model.inputs:
                x_names = [x_in.name, x_in.name.split(':')[0]]
                if x_names[0] in xs.keys():
                    xs_vals.append(xs[x_names[0]])
                    xs_names.append(x_names[0])
                elif x_names[1] in xs.keys():
                    xs_vals.append(xs[x_names[1]])
                    xs_names.append(x_names[1])
                else:
                    raise ValueError(f'Cannot map network input node {x_names[0]} to the input dict with {xs.keys()}. ')
        elif isinstance(xs, np.ndarray) or isinstance(xs, list):
            xs_vals = to_list(xs)
            assert len(model.inputs) == len(xs_vals), \
                'Number of inputs do not match the number of inputs to the functional. '
        else:
            raise TypeError('Expected a list of `np.ndarray` inputs.')

        shape_default = [x.shape for x in xs_vals]
        assert all([shape_default[0][0] == x[0] for x in shape_default[1:]])

        # prepare X,Y data.
        for i, (x, xt) in enumerate(zip(xs_vals, model.inputs)):
            x_shape = tuple(xt.get_shape().as_list())
            if x.shape[1:] != x_shape[1:]:
                try:
                    xs_vals[i] = xs_vals[i].reshape((-1,) + x_shape[1:])
                except:
                    print(
                        'Could not automatically convert the inputs to be ' 
                        'of the same size as the expected input tensors. ' 
                        'Please provide inputs of the same dimension as the `Variables`. '
                    )
                    assert False

        y_pred = to_list(model(xs_vals))

        # revert back to normal.
        # if isinstance(xs, dict):
        #     for i, (xn, xv) in enumerate(zip(xs_names, xs_vals)):
        #         xs[xn] = xv.reshpae(shape_default[i])
        if isinstance(xs, list):
            for i, x in enumerate(xs):
                xs[i] = x.reshape(shape_default[i])
        
        # return uniform shapes. 
        if all([shape_default[0]==sd for sd in shape_default[1:]]):
            try:
                y_pred = [y.reshape(shape_default[0]) for y in y_pred]
            except:
                print("Input and output dimensions need re-adjustment for post-processing.")

        return unpack_singleton(y_pred)

    @property
    def layers(self):
        return self._layers

    @layers.setter
    def layers(self, value):
        self._layers = value

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, value):
        self._inputs = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def model(self):
        self._set_model()
        return self._model

    @property
    def name(self):
        return self._layers[-1].name
    
    def _set_model(self):
        if hasattr(self, '_model'):
            if is_same_tensor(self._inputs, self._model.inputs) and \
               is_same_tensor(self._outputs, self._model.outputs):
               return
        self._model = K.function(
            unique_tensors(self._inputs),
            self._outputs
        )

    def get_weights(self, at_layer=None):
        """ Get the weights and biases of different layers.

        # Arguments
            at_layer: 
                Get the weights of a specific layer. 
            
        # Returns
            List of numpy array. 
        """
        return [l.get_weights() for l in self.layers]

    def set_weights(self, weights):
        """ Set the weights and biases of different layers.

        # Arguments
            weights: Should take the dimensions as the output of ".get_weights"
            
        # Returns 
        """
        try:
            for l, w in zip(self.layers, weights):
                l.set_weights(w)
        except:
            raise ValueError(
                'Provide data exactly the same as .get_weights() outputs. '
            )

    def count_params(self):
        """ Total number of parameters of a functional.

        # Arguments
            
        # Returns 
            Total number of parameters.
        """
        return sum([l.count_params() for l in self.layers])

    def copy(self):
        return MLPFunctional(
            inputs=self.inputs,
            outputs=self.outputs,
            layers=self.layers
        )

    def append_to_layers(self, layers):
        if self.layers is not layers:
            cl = [x.name for x in self.layers]
            for x in layers:
                if not x.name in cl:
                    self.layers += [x]

    def append_to_inputs(self, inputs):
        if self.inputs is not inputs:
            cl = [x.name for x in self.inputs]
            for x in inputs:
                if not x.name in cl:
                    self.inputs.append(x)

    def append_to_outputs(self, outputs):
        self._outputs += to_list(outputs)

    def set_trainable(self, val, layers=None):
        """ Set the weights and biases of a functional object trainable or not-trainable.
        Note: The SciModel should be called after this.

        # Arguments
            val: (Ture, False)
            layers: list of layers to be set trainable or non-trainable.
                defaulted to None.
            
        # Returns 
        """
        print("Warning: Call `model.compile()` after using set_trainable.")
        if isinstance(val, bool):
            if layers is None:
                for l in self._layers:
                    l.trainable = val
            else:
                for li in to_list(layers):
                    i = -1
                    for l in self._layers:
                        if l.count_params() > 0:
                            i += 1
                        if li == i:
                            l.trainable = val
                            break
        else:
            raise ValueError('Expected a boolean value: True or False')

    def reinitialize_weights(self):
        """ Re-initialize the weights and biases of a functional object.

        # Arguments

        # Returns 
        
        """
        for lay in self.layers:
            if hasattr(lay, 'kernel_initializer') and lay.kernel is not None:
                K.set_value(lay.kernel, lay.kernel_initializer(lay.kernel.shape))
            if hasattr(lay, 'bias_initializer') and lay.bias is not None:
                K.set_value(lay.bias, lay.bias_initializer(lay.bias.shape))

    def split(self):
        """ In the case of `Functional` with multiple outputs,
            you can split the outputs and get an associated functional.

        # Returns
            (f1, f2, ...): Tuple of splitted `Functional` objects
                associated to each output.
        """
        if len(self._outputs)==1:
            return self
        fs = ()
        # total number of outputs to get splitted.
        nr = len(self._outputs)
        # associated to each output, there is a layer to be splitted.
        lays = self.layers[:-nr]
        for out, lay in zip(self._outputs, self._layers[-nr:]):
            # copy constructor for functional.
            f = MLPFunctional(
                inputs = to_list(self.inputs),
                outputs = to_list(out),
                layers = lays + to_list(lay)
            )
            fs += (f,)
        return fs

    def __call__(self):
        return self.outputs

    def __pos__(self):
        return self

    def __neg__(self):
        return self*-1.0

    def __add__(self, other):
        return math.add(self, other)

    def __radd__(self, other):
        return math.radd(self, other)

    def __sub__(self, other):
        return math.sub(self, other)

    def __rsub__(self, other):
        return math.rsub(self, other)

    def __mul__(self, other):
        return math.mul(self, other)

    def __rmul__(self, other):
        return math.rmul(self, other)

    def __truediv__(self, other):
        return math.div(self, other)

    def __rtruediv__(self, other):
        return math.rdiv(self, other)

    def __pow__(self, power):
        return math.pow(self, power)

    def __getitem__(self, item):
        return getitem(self, item)

    def diff(self, *args, **kwargs):
        return math.diff(self, *args, **kwargs)

    def __eq__(self, other):
        return math.equal(self, other)

    def __ne__(self, other):
        return math.not_equal(self, other)

    def __gt__(self, other):
        return math.greater(self, other)

    def __ge__(self, other):
        return math.greater_equal(self, other)

    def __lt__(self, other):
        return math.less(self, other)

    def __le__(self, other):
        return math.less_equal(self, other)

    @classmethod
    def get_class(cls):
        return MLPFunctional



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..utils import default_bias_initializer, default_kernel_initializer, default_constant_initializer
from ..utils import prepare_default_activations_and_initializers
from ..utils import default_regularizer
from ..utils import floatx, set_floatx

from keras.layers import Dense
from keras.activations import linear


class Field(Dense):
    """ Configures the `Field` class for the model outputs.

    # Arguments
        name: String.
            Assigns a layer name for the output.
        units: Positive integer.
            Dimension of the output of the network.
        activation: Callable.
            A callable object for the activation.
        kernel_initializer: Initializer for the kernel.
            Defaulted to a normal distribution.
        bias_initializer: Initializer for the bias.
            Defaulted to a normal distribution.
        kernel_regularizer: Regularizer for the kernel.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        bias_regularizer: Regularizer for the bias.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        trainable: Boolean to activate parameters of the network.
        use_bias: Boolean to add bias to the network.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').

    # Raises

    """
    def __init__(self,
                 name=None,
                 units=1,
                 activation=linear,
                 kernel_initializer=None,
                 bias_initializer=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 trainable=True,
                 use_bias=True,
                 dtype=None):
        if not dtype:
            dtype = floatx()
        elif not dtype == floatx():
            set_floatx(dtype)

        assert isinstance(name, str), \
            "Please provide a string for field name. "

        # prepare initializers.
        if isinstance(kernel_initializer, (float, int)):
            kernel_initializer = default_constant_initializer(kernel_initializer)
        if isinstance(bias_initializer, (float, int)):
            bias_initializer = default_constant_initializer(bias_initializer)
        if isinstance(kernel_initializer, type(None)) and isinstance(bias_initializer, type(None)):
            activation, kernel_initializer, bias_initializer = [
                a[0] for a in prepare_default_activations_and_initializers(activation)
            ]
        # prepare regularizers.
        kernel_regularizer = default_regularizer(kernel_regularizer)
        bias_regularizer = default_regularizer(bias_regularizer)

        super(Field, self).__init__(
            units=units,
            activation=activation,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            trainable=trainable,
            use_bias=use_bias,
            name=name,
            dtype=dtype,
        )

    @staticmethod
    def prepare_field_inputs(*args):
        if len(args) == 1 and isinstance(args[0], str):
            fld_name, fld_units = args[0], 1
        elif len(args) == 2 and isinstance(args[0], str):
            fld_name, fld_units = args[0], args[1]
        elif len(args) == 2 and isinstance(args[1], str):
            fld_name, fld_units = args[1], args[0]
        else:
            raise ValueError(
                'Unrecognised inputs for output layer - please use sn.Field to custom define the outputs. '
            )
        return fld_name, fld_units